from __future__ import annotations

import numpy as np
from scipy.ndimage import gaussian_laplace
from skimage.feature import peak_local_max
from skimage.feature.blob import _prune_blobs
from skimage.morphology import ball, h_maxima
from skimage.util import img_as_float32

from faim_ipa.detection.utils import estimate_log_rescale_factor


def detect_blobs(
    img: np.ndarray,
    axial_sigma: float,
    lateral_sigma: float,
    h: int,
    scale_factors: list[int],
    overlap: float,
    background_img: np.ndarray | None = None,
    mask: np.ndarray | None = None,
) -> np.ndarray:
    """Detect blobs of different sizes.

    The blob detection finds blobs of different sizes by applying a
    Laplacian of Gaussian with increasing sigmas followed by h-maxima
    filtering. The blob detection starts with the provided sigmas and
    then doubles them for `n_scale_levels`.

    Parameters
    ----------
    img :
        Image containing spot signal.
    axial_sigma :
        Z extension of the spots.
    lateral_sigma :
        YX extension of the spots.
    h :
        h-maxima threshold.
    scale_factors :
        List of scaling factors to apply to the sigmas.
    overlap :
        A value between 0 and 1. If the fraction of area overlapping for 2
        blobs is greater than `overlap` the smaller blob is eliminated.
    background_img :
        Estimated background image. This is subtracted before the
        blob detection.
    mask :
        Foreground mask to restrict the blob detection.

    Returns
    -------
    Detected spots.
    """
    image = (
        img_as_float32(img) - img_as_float32(background_img)
        if background_img is not None
        else img_as_float32(img)
    )

    rescale_factor = estimate_log_rescale_factor(
        (axial_sigma, lateral_sigma, lateral_sigma),
    )

    sigmas = [
        (axial_sigma * f, lateral_sigma * f, lateral_sigma * f) for f in scale_factors
    ]

    scale_cube = np.empty((*image.shape, len(sigmas)), dtype=np.uint8)

    h_ = img_as_float32(np.array(h, dtype=img.dtype))
    scale_norm = np.mean([axial_sigma, lateral_sigma, lateral_sigma])
    for i, sigma in enumerate(sigmas):
        log_img = (
            -gaussian_laplace(image, sigma=sigma)
            * rescale_factor
            * (np.mean(sigma) / scale_norm) ** 2
        )
        h_detections = h_maxima(log_img, h=h_, footprint=ball(1))
        if mask is not None:
            h_detections = h_detections * mask
        scale_cube[..., i] = h_detections

    maxima = peak_local_max(
        scale_cube,
        threshold_abs=0.1,
        exclude_border=False,
        footprint=np.ones((3,) * scale_cube.ndim),
    )

    # Convert local_maxima to float64
    lm = maxima.astype(np.float64)

    # translate final column of lm, which contains the index of the
    # sigma that produced the maximum intensity value, into the sigma
    sigmas_of_peaks = np.array(sigmas)[maxima[:, -1]]

    # Remove sigma index and replace with sigmas
    lm = np.hstack([lm[:, :-1], sigmas_of_peaks])

    sigma_dim = sigmas_of_peaks.shape[1]

    if len(lm) == 0:
        return np.empty((0, 6))
    else:
        return _prune_blobs(np.array(lm), overlap=overlap, sigma_dim=sigma_dim)



from typing import Any

from ome_zarr.scale import Scaler
from ome_zarr.writer import write_image, write_multiscales_metadata
from zarr import Group


def _copy_multiscales_metadata(parent_group, subgroup):
    datasets = parent_group.attrs.asdict()["multiscales"][0]["datasets"]
    axes = parent_group.attrs.asdict()["multiscales"][0]["axes"]
    write_multiscales_metadata(subgroup, datasets=datasets, axes=axes)


def write_labels_to_group(
    labels,
    labels_name,
    parent_group: Group,
    *,
    storage_options: dict[str, Any],
    max_layer: int = 0,
    overwrite: bool = False,
):
    """
    Write labels to zarr group and copy multiscales metadata from parent group.
    """
    subgroup = parent_group.require_group(
        f"labels/{labels_name}",
        overwrite=overwrite,
    )

    axes = parent_group.attrs.asdict()["multiscales"][0]["axes"]
    if len(axes) == len(labels.shape):
        message = f"Group axes don't match label image dimensions: {len(axes)} <> {len(labels.shape)}."
        raise ValueError(message)

    write_image(
        image=labels,
        group=subgroup,
        axes=axes,
        storage_options=storage_options,
        scaler=Scaler(max_layer=max_layer),
    )

    _copy_multiscales_metadata(parent_group, subgroup)



from abc import ABC, abstractmethod
from collections.abc import Iterable
from enum import Enum
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd

from faim_ipa.io.metadata import ChannelMetadata
from faim_ipa.stitching.tile import Tile


class TileAlignmentOptions(Enum):
    """Tile alignment options."""

    STAGE_POSITION = "StageAlignment"
    GRID = "GridAlignment"


class PlateAcquisition(ABC):
    _acquisition_dir = None
    _wells = None
    _alignment: TileAlignmentOptions = None
    _background_correction_matrices: dict[str, Path | str] | None
    _illumination_correction_matrices: dict[str, Path | str] | None
    _common_well_shape: tuple[int, int, int, int, int] = None

    def __init__(
        self,
        acquisition_dir: Path | str,
        alignment: TileAlignmentOptions,
        background_correction_matrices: dict[str, Path | str] | None,
        illumination_correction_matrices: dict[str, Path | str] | None,
    ) -> None:
        self._acquisition_dir = acquisition_dir
        self._alignment = alignment
        self._background_correction_matrices = background_correction_matrices
        self._illumination_correction_matrices = illumination_correction_matrices
        self._wells = self._build_well_acquisitions(self._parse_files())
        super().__init__()

    @abstractmethod
    def _parse_files(self) -> pd.DataFrame:
        """Parse all files in the acquisition directory.

        Returns
        -------
        DataFrame
            Table of all files in the acquisition.
        """
        raise NotImplementedError

    @abstractmethod
    def _build_well_acquisitions(self, files: pd.DataFrame) -> list["WellAcquisition"]:
        """List of wells."""
        raise NotImplementedError

    def get_well_acquisitions(
        self, selection: list[str] | None = None
    ) -> list["WellAcquisition"]:
        if selection is None:
            return self._wells
        return [well for well in self._wells if well.name in selection]

    @abstractmethod
    def get_channel_metadata(self) -> dict[int, ChannelMetadata]:
        """Channel metadata."""
        raise NotImplementedError

    def get_well_names(self, wells: list[str] | None = None) -> Iterable[str]:
        """
        Get the names of all wells in the acquisition.
        """
        for well in self.get_well_acquisitions(selection=wells):
            yield well.name

    def get_omero_channel_metadata(self) -> list[dict]:
        """
        Get the channel metadata in OMERO format.

        Returns
        -------
            List of channel metadata.
        """
        ome_channels = []
        ch_metadata = self.get_channel_metadata()
        max_channel = max(list(ch_metadata.keys()))
        for index in range(max_channel + 1):
            if index in ch_metadata:
                metadata = ch_metadata[index]
                ome_channels.append(
                    {
                        "active": True,
                        "coefficient": 1,
                        "color": metadata.display_color,
                        "family": "linear",
                        "inverted": False,
                        "label": metadata.channel_name,
                        "wavelength_id": f"C{str(metadata.channel_index + 1).zfill(2)}",
                        "window": {
                            "min": np.iinfo(np.uint16).min,
                            "max": np.iinfo(np.uint16).max,
                            "start": np.iinfo(np.uint16).min,
                            "end": np.iinfo(np.uint16).max,
                        },
                    }
                )
            else:
                ome_channels.append(
                    {
                        "active": False,
                        "coefficient": 1,
                        "color": "#000000",
                        "family": "linear",
                        "inverted": False,
                        "label": "empty",
                        "wavelength_id": f"C{str(index + 1).zfill(2)}",
                        "window": {
                            "min": np.iinfo(np.uint16).min,
                            "max": np.iinfo(np.uint16).max,
                            "start": np.iinfo(np.uint16).min,
                            "end": np.iinfo(np.uint16).max,
                        },
                    }
                )

        return ome_channels

    def get_common_well_shape(self) -> tuple[int, int, int, int, int]:
        """
        Compute the maximum well extent such that each well is covered.

        Returns
        -------
            (time, channel, z, y, x)
        """
        if self._common_well_shape is None:
            well_shapes = [well.get_shape() for well in self.get_well_acquisitions()]
            self._common_well_shape = tuple(np.max(well_shapes, axis=0))

        return self._common_well_shape


class WellAcquisition(ABC):
    """
    A single well of a plate acquisition.
    """

    name: str = None
    _files = None
    _alignment: TileAlignmentOptions = None
    _background_correction_matrices: dict[str, Path | str] | None
    _illumination_correction_matrices: dict[str, Path | str] | None
    _tiles = None
    _shape: tuple[int, int] = None
    _dtype: np.dtype = None

    def __init__(
        self,
        files: pd.DataFrame,
        alignment: TileAlignmentOptions,
        background_correction_matrices: dict[str, Path | str] | None,
        illumination_correction_matrices: dict[str, Path | str] | None,
    ) -> None:
        if files["well"].nunique() != 1:
            msg = "WellAcquisition must contain files from a single well."
            raise ValueError(msg)
        self.name = files["well"].iloc[0]
        self._files = files
        self._alignment = alignment
        self._background_correction_matrices = background_correction_matrices
        self._illumination_correction_matrices = illumination_correction_matrices
        self._tiles = self._align_tiles(tiles=self._assemble_tiles())
        super().__init__()

    @abstractmethod
    def _assemble_tiles(self) -> list[Tile]:
        """Parse all tiles in the well."""
        raise NotImplementedError

    def get_dtype(self) -> np.dtype:
        """
        Get the data type of the well acquisition.

        Returns
        -------
            type
        """
        if self._dtype is None:
            self._dtype = self._tiles[0].load_data().dtype

        return self._dtype

    def _align_tiles(self, tiles: list[Tile]) -> list[Tile]:
        if self._alignment == TileAlignmentOptions.STAGE_POSITION:
            from faim_ipa.alignment import StageAlignment

            return StageAlignment(tiles=tiles).get_tiles()

        if self._alignment == TileAlignmentOptions.GRID:
            from faim_ipa.alignment import GridAlignment

            return GridAlignment(tiles=tiles).get_tiles()

        msg = f"Unknown alignment option: {self._alignment}"
        raise ValueError(msg)

    def get_tiles(self) -> list[Tile]:
        """List of tiles."""
        return self._tiles

    def get_row_col(self) -> tuple[str, str]:
        """
        Get the row and column of the well acquisition.

        Returns
        -------
            row, column
        """
        return self.name[0], self.name[1:]

    @abstractmethod
    def get_axes(self) -> list[str]:
        """
        Get the axes of the well acquisition.
        """
        raise NotImplementedError

    @abstractmethod
    def get_yx_spacing(self) -> tuple[float, float]:
        """
        Get the yx spacing of the well acquisition.
        """
        raise NotImplementedError

    @abstractmethod
    def get_z_spacing(self) -> float | None:
        """
        Get the z spacing of the well acquisition.
        """
        raise NotImplementedError

    def get_coordinate_transformations(
        self,
        max_layer: int,
        yx_binning: int,
        ndim: int,
    ) -> list[dict[str, Any]]:
        """
        Get the NGFF conform coordinate transformations for the well
        acquisition.

        Parameters
        ----------
        max_layer : Maximum layer of the resolution pyramid.
        yx_binning : Bin factor of the yx resolution.
        ndim : Number of dimensions of the data.

        Returns
        -------
            List of coordinate transformations.
        """
        transformations = []
        for s in range(max_layer + 1):
            if self.get_z_spacing() is not None:
                transformations.append(
                    [
                        {
                            "scale": [
                                1.0,
                            ]
                            * (ndim - 3)
                            + [
                                float(self.get_z_spacing()),
                                float(self.get_yx_spacing()[0] * yx_binning * 2**s),
                                float(self.get_yx_spacing()[1] * yx_binning * 2**s),
                            ],
                            "type": "scale",
                        }
                    ]
                )
            else:
                transformations.append(
                    [
                        {
                            "scale": [
                                1.0,
                            ]
                            * (ndim - 2)
                            + [
                                float(self.get_yx_spacing()[0] * yx_binning * 2**s),
                                float(self.get_yx_spacing()[1] * yx_binning * 2**s),
                            ],
                            "type": "scale",
                        }
                    ]
                )

        return transformations

    def get_shape(self):
        """
        Compute the theoretical shape of the stitched well image.
        """
        if self._shape is None:
            tile_extents = [
                tile.get_position()
                + np.array((1,) * (5 - len(tile.shape)) + tile.shape)
                for tile in self._tiles
            ]
            self._shape = tuple(np.max(tile_extents, axis=0))

        return self._shape



from __future__ import annotations

import numpy as np
from scipy.ndimage import gaussian_laplace
from skimage.morphology import ball, h_maxima, disk
from skimage.util import img_as_float32

from faim_ipa.detection.utils import (
    estimate_log_rescale_factor,
)


def detect_spots(
    img: np.ndarray,
    axial_sigma: float,
    lateral_sigma: float,
    h: int,
    background_img: np.ndarray | None = None,
    mask: np.ndarray | None = None,
) -> np.ndarray:
    """Detect diffraction limited spots.

    The spot detection uses a Laplacian of Gaussian filter to detect
    spots of a given size. These detections are intensity filtered with
    a h-maxima filter.

    Parameters
    ----------
    img :
        Image containing spot signal.
    axial_sigma :
        Z extension of the spots.
    lateral_sigma :
        YX extension of the spots.
    h :
        h-maxima threshold.
    background_img :
        Estimated background image. This is subtracted before the
        spot detection.
    mask :
     Foreground mask to restrict the spot detection.

    Returns
    -------
    Detected spots.
    """
    image = (
        img_as_float32(img) - img_as_float32(background_img)
        if background_img is not None
        else img_as_float32(img)
    )

    rescale_factor = estimate_log_rescale_factor(
        sigmas=(axial_sigma, lateral_sigma, lateral_sigma)
    )
    log_img = (
        -gaussian_laplace(image, sigma=(axial_sigma, lateral_sigma, lateral_sigma))
        * rescale_factor
    )

    h_ = img_as_float32(np.array(h, dtype=img.dtype))
    h_detections = h_maxima(log_img, h=h_, footprint=ball(1))
    if mask is not None:
        h_detections = h_detections * mask
    return np.array(np.where(h_detections)).T


def detect_spots_2d(
    img: np.ndarray,
    lateral_sigma: float,
    h: int,
    background_img: np.ndarray | None = None,
    mask: np.ndarray | None = None,
) -> np.ndarray:
    """Detect diffraction limited spots.

    The spot detection uses a Laplacian of Gaussian filter to detect
    spots of a given size. These detections are intensity filtered with
    a h-maxima filter.

    Parameters
    ----------
    img :
        Image containing spot signal.
    lateral_sigma :
        YX extension of the spots.
    h :
        h-maxima threshold.
    background_img :
        Estimated background image. This is subtracted before the
        spot detection.
    mask :
        Foreground mask to restrict the spot detection.

    Returns
    -------
    Detected spots.
    """
    image = (
        img_as_float32(img) - img_as_float32(background_img)
        if background_img is not None
        else img_as_float32(img)
    )

    rescale_factor = estimate_log_rescale_factor(sigmas=(lateral_sigma, lateral_sigma))
    log_img = (
        -gaussian_laplace(image, sigma=(lateral_sigma, lateral_sigma)) * rescale_factor
    )

    h_ = img_as_float32(np.array(h, dtype=img.dtype))
    h_detections = h_maxima(log_img, h=h_, footprint=disk(1))
    if mask is not None:
        h_detections = h_detections * mask
    return np.array(np.where(h_detections)).T



import re
import xml.etree.ElementTree as ET
from glob import glob
from os.path import basename, join
from typing import Dict, List

import numpy as np
import pandas as pd
from dateutil import parser
from tifffile import imread


def get_metadata(input_dir: str):
    """
    Extract acquisition date, pixel-size information, acquisition camera
    index, and channel information (laser, filter and objective) from a
    Yokogawa CV7000 or CV8000 acquisition. The information is extracted from
    mse- and mrf-files written by Yokogawa.

    :param input_dir: location of the Yokogawa acquisition
    :return: acquisition_date, pixel_size, pixel_size_unit, channel_information
    """
    mrf_file = join(input_dir, "MeasurementDetail.mrf")

    mrf_tree = ET.parse(mrf_file)
    mrf_root = mrf_tree.getroot()
    mrf_ns = mrf_root.tag.replace("MeasurementDetail", "")

    date_format_str = mrf_root.attrib[mrf_ns + "BeginTime"]
    date = parser.parse(date_format_str)
    acquisition_date_str = date.strftime("%Y-%m-%d")

    channels = {}
    for child in mrf_root:
        ch = child.get(mrf_ns + "Ch")
        if ch is not None:
            channels[ch] = {
                "pixel_size": child.get(mrf_ns + "HorizontalPixelDimension"),
                "cam_index": child.get(mrf_ns + "CameraNumber"),
            }

    pixel_size = float(mrf_root[1].attrib.get(mrf_ns + "HorizontalPixelDimension"))
    pixel_size_unit = "micron"

    mes_file = glob(join(input_dir, "*.mes"))[0]

    mes_tree = ET.parse(mes_file)
    mes_root = mes_tree.getroot()
    mes_ns = mes_root.tag.replace("MeasurementSetting", "")

    for child in mes_root[2]:
        ch = child.get(mes_ns + "Ch")
        if ch is not None:
            channel_dict = channels[ch]
            channel_dict["objective"] = child.get(mes_ns + "Objective").replace(
                " ", "-"
            )
            channel_dict["filter"] = child.get(mes_ns + "Acquisition").replace("/", "-")
            channel_dict["laser"] = child[0].text

    return acquisition_date_str, pixel_size, pixel_size_unit, channels


def parse_filename(file_name: str, plate_name: str):
    """
    Parse CellVoyager image file names.
    File name pattern:
    {plate_name}_{well}_{timepoint}{field}{L}{action}{Z-plane}{channel}.tif

    :param file_name:
    :param plate_name:
    :return: plate_name, well, time_point, field, line, action, z, channel
    """
    tmp = basename(file_name)
    tmp = tmp.replace(plate_name, "")

    regex = re.compile(
        "_([A-Z]{1}[0-9]{2})_(T[0-9]{4})F([0-9]+)(L[0-9]{2})(A[0-9]+)Z([0-9]+)(C[0-9]{2}).tif"
    )

    well, time_point, field, line, action, z, channel = regex.findall(tmp)[0]

    return plate_name, well, time_point, int(field), line, action, int(z), channel


def create_table(files: List[str], plate_name: str) -> pd.DataFrame:
    """
    Create table of file-names with columns for plate, well, time-point,
    field, L, action, Z and channel.

    :param files: image file list
    :param plate_name: Name of the plate
    :return: table
    """
    plate = []
    well = []
    timepoint = []
    field = []
    lines = []
    action = []
    z = []
    channel = []
    path = []
    for file in files:
        p, w, t, f, line, a, z_, c = parse_filename(file, plate_name)
        plate.append(p)
        well.append(w)
        timepoint.append(t)
        field.append(f)
        lines.append(line)
        action.append(a)
        z.append(z_)
        channel.append(c)
        path.append(file)

    return pd.DataFrame(
        {
            "plate": plate,
            "well": well,
            "timepoint": timepoint,
            "field": field,
            "line": lines,
            "action": action,
            "z": z,
            "channel": channel,
            "path": path,
        }
    )


def build_field_stacks_for_channels(table: pd.DataFrame, z_plane: int) -> Dict:
    """
    Stack all fields of a given z-plane for every channel and return them.

    :param table: of all image files
    :param z_plane: to stack
    :return: field stack for each found channel
    """
    z_table = table[table["z"] == z_plane]
    acquired_channels = z_table.channel.unique().tolist()
    acquired_channels.sort()

    stacks = {}
    for ch in acquired_channels:
        sub_table = z_table[z_table["channel"] == ch]
        fields = []
        for file_path in sub_table["path"]:
            fields.append(imread(file_path))

        stacks[ch] = np.array(fields)

    return stacks


def subtract_dark_images(stacks: Dict, channel_metadata: Dict, input_dir: str) -> Dict:
    """
    Subtract camera dark images from channels.

    :param stacks: field stacks for each channel
    :param channel_metadata: metadata about channels
    :param input_dir: location of the Yokogawa acquisition
    :return: dark image subtracted stacks
    """
    dark_img_subtracted = {}
    for key, value in stacks.items():
        channel_idx = str(int(key[1:]))
        cam_idx = channel_metadata[channel_idx]["cam_index"]
        dark_img = imread(glob(join(input_dir, f"*_CAM{cam_idx}.tif"))[0])

        dark_img_subtracted[key] = value - dark_img

    return dark_img_subtracted


def compute_median_projection(stacks: Dict) -> Dict:
    """
    Compute median projection of the fields for each channel.

    :param stacks: field stacks
    :return: median projection for each channel
    """
    projections = {}
    for key, value in stacks.items():
        projections[key] = np.median(value, axis=0)

    return projections


def get_output_name(acquisition_date: str, channel: Dict) -> str:
    """
    Create output name for reference images.

    :param acquisition_date:
    :param channel:
    :return: "{acquisition_date}_{objective}_{laser}_{filter}.tif"
    """
    obj = channel["objective"]
    laser = channel["laser"]
    filter = channel["filter"]
    return f"{acquisition_date}_{obj}_{laser}_{filter}.tif"



import subprocess
import sys
from inspect import cleandoc


def git_status_clean(src_dir: str = "src"):
    command = ["git", "status", "--porcelain", src_dir]
    info = subprocess.run(command, capture_output=True, text=True, check=False)
    changes = len(info.stdout.splitlines())

    if changes > 0:
        sys.tracebacklimit = 0
        message = cleandoc(
            f"""
        There are {changes} untracked changes in {src_dir}.

        Please commit or stash before proceeding.
        """
        )
        raise RuntimeError(message)
    sys.exit(0)


if __name__ == "__main__":
    if len(sys.argv) > 1:
        git_status_clean(src_dir=str(sys.argv[1]))
    git_status_clean()



import numpy as np
from scipy.ndimage import gaussian_filter, gaussian_laplace


def compute_axial_sigma(
    wavelength: float, NA: float, axial_spacing: float  # noqa: N803
):
    """
    Sigma which produces a Gaussian with the same full width
    half maximum as described by Abbe's diffraction formula for axial resolution.

    R = 2 * lambda / (NA**2)

    Parameters
    ----------
    wavelength :
        Emission wavelength
    NA :
        Numerical Aperture
    axial_spacing :
        Spacing of z planes in nanometers

    Returns
    -------
    theoretical sigma in pixels
    """
    return 2 * wavelength / (NA**2) / (2 * np.sqrt(2 * np.log(2))) / axial_spacing


def compute_lateral_sigma(
    wavelength: float, NA: float, lateral_spacing: float  # noqa: N803
):
    """
    Sigma which produces a Gaussian with the same full width
    half maximum as the theoretical resolution limit in Y/X described by E. Abbe.

    d = lambda / (2*NA)

    Parameters
    ----------
    wavelength :
        Emission wavelength
    NA :
        Numerical Aperture
    lateral_spacing :
        Pixel size in YX.

    Returns
    -------
    theoretical sigma in pixels
    """
    return wavelength / (2 * NA) / (2 * np.sqrt(2 * np.log(2))) / lateral_spacing


def estimate_log_rescale_factor(sigmas: tuple[float, ...]) -> float:
    """
    Estimate the rescale factor for a LoG filter response, such that
    the LoG filter response intensities are equal to the input image
    intensities for spots of size equal to a Gaussian with sigmas
    (axial_sigma, lateral_sigma, lateral_sigma).

    Note: Arbitrary number of sigmas is possible.

    Parameters
    ----------
    sigmas :
        Tuple of sigmas used in LoG detection.

    Returns
    -------
    rescale_factor
    """
    extend = int(max(sigmas) * 7)
    img = np.zeros((extend,) * len(sigmas), dtype=np.float32)
    img[(extend // 2,) * len(sigmas)] = 1
    img = gaussian_filter(img, sigmas)
    img = img / img.max()
    img_log = -gaussian_laplace(input=img, sigma=sigmas)
    return 1 / img_log.max()



from __future__ import annotations

import numpy as np


def mean_cast_to(target_dtype):
    """
    Wrap np.mean to cast the result to a given dtype.
    """

    def _mean(
        a,
        axis=None,
        dtype=None,
        out=None,
        keepdims=np._NoValue,
        *,
        where=np._NoValue,
    ):
        return np.mean(
            a=a, axis=axis, dtype=dtype, out=out, keepdims=keepdims, where=where
        ).astype(target_dtype)

    return _mean



from copy import copy

import numpy as np
from numpy._typing import NDArray
from scipy.ndimage import distance_transform_cdt

from faim_ipa.stitching.tile import Tile, TilePosition


def fuse_linear(warped_tiles: NDArray, warped_distance_masks: NDArray) -> NDArray:
    """
    Fuse transformed tiles using a linear gradient to compute the weighted
    average where tiles are overlapping.

    Parameters
    ----------
    warped_tiles :
        Tile images transformed to the final image space.
    warped_distance_masks :
        Distance masks for the transformed tiles. They are non-zero for
        foreground pixels, and the value is the distance to the closest edge of
        the tile.

    Returns
    -------
    Fused image.
    """
    dtype = warped_tiles.dtype
    fused_image = np.zeros_like(warped_tiles[0], dtype=np.float32)

    if warped_tiles.shape[0] > 1:
        denominator = np.sum(warped_distance_masks, axis=0)

        for tile, mask in zip(warped_tiles, warped_distance_masks, strict=True):
            weight = np.divide(mask, denominator, where=denominator > 0)
            np.clip(weight, 0, 1, out=weight)
            np.add(
                fused_image,
                tile.astype(np.float32) * weight,
                out=fused_image,
                where=weight > 0,
            )
    else:
        np.add(
            fused_image,
            warped_tiles[0],
            out=fused_image,
            where=warped_distance_masks[0] > 0,
        )

    return np.clip(fused_image, 0, np.iinfo(dtype).max).astype(dtype)


def fuse_linear_random(
    warped_tiles: NDArray, warped_distance_masks: NDArray
) -> NDArray:
    """
    Fuse transformed tiles by sampling random pixels where tiles are
    overlapping, using a linear gradient to compute the random weights.

    Parameters
    ----------
    warped_tiles :
        Tile images transformed to the final image space.
    warped_distance_masks :
        Distance masks for the transformed tiles. They are non-zero for
        foreground pixels, and the value is the distance to the closest edge of
        the tile.

    Returns
    -------
    Fused image.
    """
    np.random.seed(0)
    dtype = warped_tiles.dtype
    if warped_tiles.shape[0] > 1:
        denominator = warped_distance_masks.sum(axis=0)
        weights = np.true_divide(
            warped_distance_masks, denominator, where=denominator > 0
        )
        weights = np.clip(np.nan_to_num(weights, nan=0, posinf=1, neginf=0), 0, 1)
        weights = np.cumsum(weights, axis=0)
        weights = np.insert(weights, 0, np.zeros_like(weights[0]), axis=0)
        rand_tile = np.random.rand(*warped_tiles.shape[1:])
        for i in range(len(warped_tiles)):
            warped_tiles[i, (rand_tile < weights[i]) | (weights[i + 1] < rand_tile)] = 0

    return np.sum(warped_tiles, axis=0).astype(dtype)


def fuse_mean(warped_tiles: NDArray, warped_distance_masks: NDArray) -> NDArray:
    """
    Fuse transformed tiles and compute the mean of the overlapping pixels.

    Parameters
    ----------
    warped_tiles :
        Tile images transformed to the final image space.
    warped_distance_masks :
        Distance masks for the transformed tiles. They are non-zero for
        foreground pixels, and the value is the distance to the closest edge of
        the tile.

    Returns
    -------
    Fused image.
    """
    denominator = np.sum(warped_distance_masks > 0, axis=0)
    fused_image = np.zeros_like(warped_tiles[0], dtype=np.float32)

    for tile, mask in zip(warped_tiles, warped_distance_masks, strict=True):
        weight = np.divide(mask > 0, denominator, where=denominator > 0)
        np.add(
            fused_image,
            tile.astype(np.float32) * weight,
            out=fused_image,
            where=weight > 0,
        )

    return np.clip(fused_image, 0, np.iinfo(warped_tiles.dtype).max).astype(
        warped_tiles.dtype
    )


def fuse_sum(
    warped_tiles: NDArray, warped_distance_masks: NDArray  # noqa: ARG001
) -> NDArray:
    """
    Fuse transformed tiles and compute the sum of the overlapping pixels.

    Parameters
    ----------
    warped_tiles :
        Tile images transformed to the final image space.
    warped_distance_masks :
        Distance masks for the transformed tiles. They are non-zero for
        foreground pixels, and the value is the distance to the closest edge of
        the tile. (Not used in this function)

    Returns
    -------
    Fused image.
    """
    fused_image = np.sum(warped_tiles, axis=0)
    return fused_image.astype(warped_tiles.dtype)


def fuse_overlay_fwd(warped_tiles: NDArray, warped_distance_masks: NDArray) -> NDArray:
    """
    Fuse transformed tiles. Where tiles overlap, the tile later in the sequence
    overwrites the earlier one.

    Parameters
    ----------
    warped_tiles :
        Tile images transformed to the final image space.
    warped_distance_masks :
        Distance masks for the transformed tiles. They are non-zero for
        foreground pixels, and the value is the distance to the closest edge of
        the tile.

    Returns
    -------
    Fused image.
    """

    warped_masks = warped_distance_masks.astype(bool)

    fused_image = np.zeros_like(warped_tiles[0])
    for tile, mask in zip(warped_tiles, warped_masks, strict=True):
        fused_image[mask] = tile[mask]

    return fused_image


def fuse_overlay_bwd(warped_tiles: NDArray, warped_distance_masks: NDArray) -> NDArray:
    """
    Fuse transformed tiles. Where tiles overlap, the tile earlier in the
    sequence overwrites the later one.

    Parameters
    ----------
    warped_tiles :
        Tile images transformed to the final image space.
    warped_distance_masks :
        Distance masks for the transformed tiles. They are non-zero for
        foreground pixels, and the value is the distance to the closest edge of
        the tile.

    Returns
    -------
    Fused image.
    """

    warped_masks = warped_distance_masks.astype(bool)

    fused_image = np.zeros_like(warped_tiles[0])
    for tile, mask in zip(reversed(warped_tiles), reversed(warped_masks), strict=True):
        fused_image[mask] = tile[mask]

    return fused_image


def translate_tiles_2d(
    block_info, chunk_shape, tiles, *, build_acquisition_mask: bool = False
):
    """
    Translate tiles to their relative position inside the given block.

    Parameters
    ----------
    block_info :
        da.map_blocks block_info.
    chunk_shape :
        shape of the chunk in zyx.
    tiles :
        list of tiles.

    Returns
    -------
        translated tiles, translated distance masks
    """
    array_location = block_info[None]["array-location"]
    chunk_zyx_origin = np.array(
        [array_location[2][0], array_location[3][0], array_location[4][0]]
    )

    if not all(tile.shape == tiles[0].shape for tile in tiles):
        msg = "All tiles must have the same shape."
        raise ValueError(msg)
    distance_mask = get_distance_mask(tiles[0].shape)

    warped_tiles = []
    warped_distance_masks = []
    for tile in tiles:
        tile_origin = np.array(tile.get_zyx_position())
        tile_data = (
            tile.load_data_mask() if build_acquisition_mask else tile.load_data()
        )
        if tile_data.ndim == 2:
            tile_data = tile_data[np.newaxis, ...]
        warped_tile = shift_yx(chunk_zyx_origin, tile_data, tile_origin, chunk_shape)
        warped_distance_mask = shift_yx(
            chunk_zyx_origin, distance_mask, tile_origin, chunk_shape
        )

        warped_tiles.append(warped_tile)
        warped_distance_masks.append(warped_distance_mask)

    return np.array(warped_tiles), np.array(warped_distance_masks)


def get_distance_mask(tile_shape):
    mask = np.zeros(tile_shape[-2:], dtype=bool)
    mask[1:-1, 1:-1] = True
    return distance_transform_cdt(mask, metric="taxicab").astype(np.uint16) + 1


def shift_yx(chunk_zyx_origin, tile_data, tile_origin, chunk_shape):
    warped_tile = np.zeros(chunk_shape, dtype=tile_data.dtype)
    yx_shift = (tile_origin - chunk_zyx_origin)[1:]
    if yx_shift[0] < 0:
        tile_start_y = abs(yx_shift[0])
        tile_end_y = min(tile_start_y + chunk_shape[1], tile_data.shape[-2])
    else:
        tile_start_y = 0
        tile_end_y = max(
            0, min(tile_start_y + chunk_shape[1] - yx_shift[0], tile_data.shape[-2])
        )
    if yx_shift[1] < 0:
        tile_start_x = abs(yx_shift[1])
        tile_end_x = min(tile_start_x + chunk_shape[2], tile_data.shape[-1])
    else:
        tile_start_x = 0
        tile_end_x = min(
            tile_start_x + chunk_shape[2] - yx_shift[1], tile_data.shape[-1]
        )
    tile_data = tile_data[..., tile_start_y:tile_end_y, tile_start_x:tile_end_x]
    if tile_data.size > 0:
        start_y = max(0, yx_shift[0])
        end_y = start_y + tile_data.shape[-2]
        start_x = max(0, yx_shift[1])
        end_x = start_x + tile_data.shape[-1]
        warped_tile[: tile_data.shape[0], start_y:end_y, start_x:end_x] = tile_data
    return warped_tile


def assemble_chunk(
    block_info=None,
    tile_map=None,
    warp_func=None,
    fuse_func=None,
    dtype=None,
    *,
    build_acquisition_mask: bool = False,
):
    """
    Assemble a chunk of the stitched image.

    Parameters
    ----------
    block_info :
        da.map_blocks block_info.
    tile_map :
        map of block positions to tiles.
    warp_func :
        function used to warp tiles.
    fuse_func :
        function used to fuse tiles.
    dtype :
        tile data type.

    Returns
    -------
        fused tiles corresponding to this block/chunk
    """
    chunk_location = block_info[None]["chunk-location"]
    chunk_shape = block_info[None]["chunk-shape"]
    tiles = tile_map[chunk_location]

    if len(tiles) > 0:
        warped_tiles, warped_distance_masks = warp_func(
            block_info,
            chunk_shape[-3:],
            tiles,
            build_acquisition_mask=build_acquisition_mask,
        )

        if len(tiles) > 1:
            stitched_img = fuse_func(
                warped_tiles,
                warped_distance_masks,
            )
            stitched_img = stitched_img[np.newaxis, np.newaxis, ...]
        else:
            stitched_img = warped_tiles[np.newaxis, ...]
    else:
        stitched_img = np.zeros(chunk_shape, dtype=dtype)

    return stitched_img


def shift_to_origin(tiles: list[Tile]) -> list[Tile]:
    """
    Shift tile positions such that the minimal position is (0, 0, 0, 0, 0).

    Parameters
    ----------
    tiles :
        List of tiles.

    Returns
    -------
    List of shifted tiles.
    """
    min_tile_origin = np.min([np.array(tile.get_position()) for tile in tiles], axis=0)
    shifted_tiles = copy(tiles)
    for tile in shifted_tiles:
        shifted_pos = np.array(tile.get_position()) - min_tile_origin
        tile.position = TilePosition(
            time=shifted_pos[0],
            channel=shifted_pos[1],
            z=shifted_pos[2],
            y=shifted_pos[3],
            x=shifted_pos[4],
        )
    return shifted_tiles



import numpy as np


def polynomial_fit(mip, polynomial_degree=3, order=None):
    """
    Fits a polynomial f(x, y) to the provided mip.

    :param mip: to fit
    :param polynomial_degree: highest degree of individual coefficients
    :param order: polynomial order
    :return: fitted matrix, string describing the polynomial
    """
    Y, X = np.meshgrid(
        np.linspace(0, 1, mip.shape[0]), np.linspace(0, 1, mip.shape[1]), indexing="ij"
    )

    coeffs = np.ones((polynomial_degree + 1, polynomial_degree + 1))

    A = np.zeros((coeffs.size, X.size))

    poly_str = ["1"]
    for row_idx, (i, j) in enumerate(np.ndindex(coeffs.shape)):
        if order is not None and i + j > order:
            tmp = np.zeros_like(X)
        else:
            tmp = coeffs[i, j] * X**j * Y**i

            substr = []
            if j > 0:
                substr.append(f"X^{j}")
            if i > 0:
                substr.append(f"Y^{i}")

            if len(substr) > 0:
                poly_str.append(" * ".join(substr))
        A[row_idx] = tmp.ravel()

    poly_str = " + ".join(poly_str)

    x, residuals, rank, s = np.linalg.lstsq(A.T, np.ravel(mip), rcond=None)

    x = x.reshape((polynomial_degree + 1, polynomial_degree + 1))
    fit = np.polynomial.polynomial.polyval2d(Y, X, x).reshape(mip.shape)
    return fit, poly_str



from enum import IntEnum


class PlateLayout(IntEnum):
    """Plate layout, 18, 24, 96 or 384-well."""

    I18 = 18
    I24 = 24
    I96 = 96
    I384 = 384


def get_rows_and_columns(layout: PlateLayout | int) -> tuple[list[str], list[str]]:
    """Return rows and columns for requested layout."""
    if layout == PlateLayout.I18:
        rows = list("ABC")
        cols = [str(i).zfill(2) for i in range(1, 7)]
    elif layout == PlateLayout.I24:
        rows = list("ABCD")
        cols = [str(i).zfill(2) for i in range(1, 7)]
    elif layout == PlateLayout.I96:
        rows = list("ABCDEFGH")
        cols = [str(i).zfill(2) for i in range(1, 13)]
    elif layout == PlateLayout.I384:
        rows = list("ABCDEFGHIJKLMNOP")
        cols = [str(i).zfill(2) for i in range(1, 25)]
    else:
        msg = f"{layout} layout not supported."
        raise NotImplementedError(msg)

    assert len(rows) * len(cols) == layout
    return rows, cols



from __future__ import annotations

import re
from copy import copy
from decimal import Decimal
from os.path import exists, join
from typing import TYPE_CHECKING
from warnings import warn

import numpy as np
import pandas as pd
from defusedxml.ElementTree import parse
from tifffile import imread
from tqdm import tqdm

from faim_ipa.hcs.acquisition import (
    PlateAcquisition,
    TileAlignmentOptions,
    WellAcquisition,
)
from faim_ipa.hcs.cellvoyager.tile import StackedTile
from faim_ipa.io.metadata import ChannelMetadata
from faim_ipa.stitching.tile import Tile, TilePosition

if TYPE_CHECKING:
    from pathlib import Path

BTS_NS = "{http://www.yokogawa.co.jp/BTS/BTSSchema/1.0}"


class CellVoyagerWellAcquisition(WellAcquisition):
    """
    Data structure for a CellVoyager well acquisition.
    """

    def __init__(
        self,
        files: pd.DataFrame,
        alignment: TileAlignmentOptions,
        metadata: pd.DataFrame,
        background_correction_matrices: dict[str, Path | str] | None = None,
        illumination_correction_matrices: dict[str, Path | str] | None = None,
        n_planes_in_stacked_tile: int = 1,
        dtype: np.dtype | None = None,
    ):
        self._metadata = metadata
        self._z_spacing = self._compute_z_spacing(files)
        self._dtype = dtype if dtype else self._get_dtype(files)
        self._n_planes_in_stacked_tile = n_planes_in_stacked_tile
        super().__init__(
            files=files,
            alignment=alignment,
            background_correction_matrices=background_correction_matrices,
            illumination_correction_matrices=illumination_correction_matrices,
        )

    def _get_dtype(self, files: pd.DataFrame) -> np.dtype:
        return imread(files["path"].iloc[0]).dtype

    def _compute_z_spacing(self, files: pd.DataFrame) -> float | None:
        if "ZIndex" in files.columns:
            z_steps = np.array(
                files.astype({"Z": float}).groupby("ZIndex", sort=True).mean("Z")["Z"]
            )

            precision = -Decimal(str(z_steps[0])).as_tuple().exponent
            return np.round(np.mean(np.diff(z_steps)), decimals=precision)
        return None

    def _assemble_tiles(self) -> list[Tile]:
        min_z_index = 0
        max_z_index = min_z_index + 1
        if "ZIndex" in self._files.columns:
            min_z_index = self._files["ZIndex"].min()
            max_z_index = self._files["ZIndex"].max() + 1

        tiles = {}
        for _i, row in self._files.iterrows():
            if "ZIndex" in row:
                tile_z_index = (
                    row["ZIndex"] - min_z_index
                ) // self._n_planes_in_stacked_tile
            else:
                tile_z_index = min_z_index
            tczyx_index = (
                row["TimePoint"],
                row["Ch"],
                tile_z_index,
                row["Y"],
                row["X"],
            )
            if tczyx_index not in tiles:
                tiles[tczyx_index] = [row]
            else:
                tiles[tczyx_index].append(row)

        stacked_tiles = []
        for tczyx_index, rows in tiles.items():
            row_dict = {}
            for r in rows:
                if "ZIndex" in r:
                    row_dict[r["ZIndex"]] = r["path"]
                else:
                    row_dict[min_z_index] = r["path"]

            files = []
            z_start = tczyx_index[2] * self._n_planes_in_stacked_tile + min_z_index
            for z in range(
                z_start, min(z_start + self._n_planes_in_stacked_tile, max_z_index)
            ):
                if z in row_dict:
                    files.append(row_dict[z])
                else:
                    files.append(None)

            time_point = tczyx_index[0]
            channel = tczyx_index[1]
            y, x = tczyx_index[3], tczyx_index[4]

            ch_metadata = self._metadata[self._metadata["Ch"] == channel].iloc[0]
            shape = (
                len(files),
                int(ch_metadata["VerticalPixels"]),
                int(ch_metadata["HorizontalPixels"]),
            )

            yx_spacing = self.get_yx_spacing()

            bgcm = None
            if self._background_correction_matrices is not None:
                bgcm = self._background_correction_matrices[str(channel)]

            icm = None
            if self._illumination_correction_matrices is not None:
                icm = self._illumination_correction_matrices[str(channel)]

            stacked_tiles.append(
                StackedTile(
                    paths=files,
                    shape=shape,
                    position=TilePosition(
                        time=time_point,
                        channel=int(channel),
                        z=z_start,
                        y=int(-float(y) / yx_spacing[0]),
                        x=int(float(x) / yx_spacing[1]),
                    ),
                    background_correction_matrix_path=bgcm,
                    illumination_correction_matrix_path=icm,
                    dtype=self._dtype,
                )
            )

        return stacked_tiles

    def get_axes(self) -> list[str]:
        if self._z_spacing is not None:
            return ["c", "z", "y", "x"]
        return ["c", "y", "x"]

    def get_yx_spacing(self) -> tuple[float, float]:
        ch_metadata = self._metadata.iloc[0]
        return (
            float(ch_metadata["VerticalPixelDimension"]),
            float(ch_metadata["HorizontalPixelDimension"]),
        )

    def get_z_spacing(self) -> float | None:
        return self._z_spacing


class StackAcquisition(PlateAcquisition):
    def __init__(
        self,
        acquisition_dir: Path | str,
        alignment: TileAlignmentOptions,
        background_correction_matrices: dict[str, Path | str] | None = None,
        illumination_correction_matrices: dict[str, Path | str] | None = None,
        n_planes_in_stacked_tile: int = 1,
    ):
        self._n_planes_in_stacked_tile = n_planes_in_stacked_tile
        super().__init__(
            acquisition_dir=acquisition_dir,
            alignment=alignment,
            background_correction_matrices=background_correction_matrices,
            illumination_correction_matrices=illumination_correction_matrices,
        )

    def get_channel_metadata(self) -> dict[int, ChannelMetadata]:
        metadata = self._parse_metadata()
        ch_metadata = {}

        for _i, row in metadata.iterrows():
            index = int(row["Ch"]) - 1
            ch_metadata[index] = ChannelMetadata(
                channel_index=index,
                channel_name=row["Ch"],
                display_color=row["Color"],
                spatial_calibration_x=row["HorizontalPixelDimension"],
                spatial_calibration_y=row["VerticalPixelDimension"],
                spatial_calibration_units="um",
                z_spacing=self.get_z_spacing(),
                wavelength=self.__parse_filter_wavelength(row["Acquisition"]),
                exposure_time=row["ExposureTime"],
                exposure_time_unit="ms",
                objective=row["Objective"],
            )

        assert min(ch_metadata.keys()) == 0, "Channel indices must start at 0."

        return ch_metadata

    def get_z_spacing(self) -> float:
        return self._wells[0].get_z_spacing()

    def _build_well_acquisitions(self, files: pd.DataFrame) -> list[WellAcquisition]:
        dtype = copy(imread(files.iloc[0]["path"]).dtype)
        return [
            CellVoyagerWellAcquisition(
                files=files[files["well"] == well],
                alignment=self._alignment,
                metadata=self._parse_metadata(),
                background_correction_matrices=self._background_correction_matrices,
                illumination_correction_matrices=self._illumination_correction_matrices,
                n_planes_in_stacked_tile=self._n_planes_in_stacked_tile,
                dtype=dtype,
            )
            for well in tqdm(files["well"].unique())
        ]

    @staticmethod
    def __parse_filter_wavelength(value) -> int:
        try:
            return int(re.match(r"BP(\d+)/", value).group(1))
        except AttributeError:  # no cov
            return 0

    def _parse_metadata(self) -> pd.DataFrame:
        mrf_file = join(self._acquisition_dir, "MeasurementDetail.mrf")
        if not exists(mrf_file):
            msg = f"MeasurementDetail.mrf not found in: {self._acquisition_dir}"
            raise ValueError(msg)
        mrf_tree = parse(mrf_file)
        mrf_root = mrf_tree.getroot()

        channels = []
        for channel in mrf_root.findall(BTS_NS + "MeasurementChannel"):
            row = {
                key.replace(BTS_NS, ""): value for key, value in channel.attrib.items()
            }
            channels.append(row)

        mes_file = join(
            self._acquisition_dir,
            mrf_root.attrib[BTS_NS + "MeasurementSettingFileName"],
        )
        if not exists(mes_file):
            msg = f"Settings file not found: {mes_file}"
            raise ValueError(msg)
        mes_tree = parse(mes_file)
        mes_root = mes_tree.getroot()

        channel_settings = []
        for channel in mes_root.find(BTS_NS + "ChannelList").findall(
            BTS_NS + "Channel"
        ):
            row = {
                key.replace(BTS_NS, ""): value for key, value in channel.attrib.items()
            }
            channel_settings.append(row)

        return pd.merge(
            pd.DataFrame(channels),
            pd.DataFrame(channel_settings),
            left_on="Ch",
            right_on="Ch",
        )

    def _parse_files(self) -> pd.DataFrame:
        mlf_file = join(self._acquisition_dir, "MeasurementData.mlf")
        if not exists(mlf_file):
            msg = f"MeasurementData.mlf not found in: {self._acquisition_dir}"
            raise ValueError(msg)
        mlf_tree = parse(mlf_file)
        mlf_root = mlf_tree.getroot()

        files = []
        for record in mlf_root.findall(BTS_NS + "MeasurementRecord"):
            row = {
                key.replace(BTS_NS, ""): value for key, value in record.attrib.items()
            }
            if row.pop("Type") == "IMG":
                row |= {
                    "path": join(self._acquisition_dir, record.text),
                    "well": chr(ord("@") + int(row.pop("Row")))
                    + row.pop("Column").zfill(2),
                }
                files.append(row)

            record.clear()

        files = pd.DataFrame(files)
        files["TimePoint"] = files["TimePoint"].astype(int)
        files["ZIndex"] = files["ZIndex"].astype(int)
        return files


class ZAdjustedStackAcquisition(StackAcquisition):
    _trace_log_files: list[str | Path]

    def __init__(
        self,
        acquisition_dir: Path | str,
        trace_log_files: list[Path | str],
        alignment: TileAlignmentOptions,
        background_correction_matrices: dict[str, Path | str] | None = None,
        illumination_correction_matrices: dict[str, Path | str] | None = None,
        n_planes_in_stacked_tile: int = 1,
    ):
        self._trace_log_files = trace_log_files
        super().__init__(
            acquisition_dir,
            alignment,
            background_correction_matrices,
            illumination_correction_matrices,
            n_planes_in_stacked_tile=n_planes_in_stacked_tile,
        )

    def _parse_files(self) -> pd.DataFrame:
        files = super()._parse_files()
        z_mapping = self.create_z_mapping(self._trace_log_files, self._acquisition_dir)
        # merge files left with mapping on path
        merged = files.merge(z_mapping, how="left", left_on=["path"], right_on=["path"])
        if np.any(merged["z_pos"].isna()):
            msg = "At least one invalid z position."
            raise ValueError(msg)
        min_z = np.min(merged["z_pos"].astype(float))
        z_spacing = np.mean(
            merged[merged["ZIndex"].astype(int) == 2]["Z"].astype(float)
        ) - np.mean(merged[merged["ZIndex"].astype(int) == 1]["Z"].astype(float))
        # Shift ZIndex for each field in each well according to the auto-focus value
        for key, selection in tqdm(
            merged.groupby(["well", "FieldIndex", "Ch"]), desc="Adjust Z", leave=False
        ):
            z_index_offset = int(
                np.round((np.min(selection["z_pos"].astype(float)) - min_z) / z_spacing)
            )
            merged.loc[
                (merged["well"] == key[0])
                & (merged["FieldIndex"] == key[1])
                & (merged["Ch"] == key[2]),
                "ZIndex",
            ] += z_index_offset

        # Start at 0
        merged["ZIndex"] = merged["ZIndex"] - merged["ZIndex"].min()
        # update Z
        merged["Z"] = merged["z_pos"]
        return merged

    @staticmethod
    def create_z_mapping(
        trace_log_files: list[str | Path], acquisition_dir: Path | str | None = None
    ) -> pd.DataFrame:
        z_pos = []
        filenames = []
        missing = []
        value = None
        for trace_file in trace_log_files:
            with open(trace_file) as log:
                for line in log:
                    tokens = line.split(",")
                    if (
                        (len(tokens) > 14)
                        and (tokens[7] == "--->")
                        and (tokens[8] == "MS_MANU")
                    ):
                        value = float(tokens[14])
                    elif (
                        (len(tokens) > 12)
                        and (tokens[7] == "--->")
                        and (tokens[8] == "AF_MANU")
                        and (tokens[9] == "34")
                    ):
                        value = float(tokens[12])
                    elif (
                        (len(tokens) > 8)
                        and (tokens[4] == "Measurement")
                        and (tokens[7] == "_init_frame_save")
                    ):
                        filename = tokens[8]
                        if value is None:
                            missing.append(
                                join(acquisition_dir, filename)
                                if acquisition_dir
                                else filename
                            )
                        else:
                            filenames.append(
                                join(acquisition_dir, filename)
                                if acquisition_dir
                                else filename
                            )
                            z_pos.append(value)
                    elif (
                        (len(tokens) > 7)
                        and (tokens[6] == "EndPeriod")
                        and (tokens[7] == "acquire frames")
                    ):
                        value = None

        if len(missing) > 0:
            warn("Z position information missing for some files.", stacklevel=2)
            warn(
                f"First file without z position information: {missing[0]}", stacklevel=2
            )
        return pd.DataFrame(
            {
                "path": filenames,
                "z_pos": z_pos,
            }
        )



from faim_ipa.hcs.imagexpress.acquisition import (
    ImageXpressPlateAcquisition,
    ImageXpressWellAcquisition,
    MixedAcquisition,
    SinglePlaneAcquisition,
    StackAcquisition,
)

__all__ = [
    "ImageXpressPlateAcquisition",
    "ImageXpressWellAcquisition",
    "MixedAcquisition",
    "SinglePlaneAcquisition",
    "StackAcquisition",
]



import os
import re
from abc import abstractmethod
from decimal import Decimal
from pathlib import Path

import numpy as np
import pandas as pd
from numpy._typing import NDArray
from tqdm import tqdm

from faim_ipa.hcs.acquisition import (
    PlateAcquisition,
    TileAlignmentOptions,
    WellAcquisition,
)
from faim_ipa.io.metadata import ChannelMetadata
from faim_ipa.io.metaseries import load_metaseries_tiff_metadata
from faim_ipa.stitching.tile import Tile, TilePosition
from faim_ipa.utils import rgb_to_hex, wavelength_to_rgb


class ImageXpressWellAcquisition(WellAcquisition):
    def __init__(
        self,
        files: pd.DataFrame,
        alignment: TileAlignmentOptions,
        z_spacing: float | None,
        background_correction_matrices: dict[str, Path | str] | None = None,
        illumination_correction_matrices: dict[str, Path | str] | None = None,
    ) -> None:
        self._z_spacing = z_spacing
        super().__init__(
            files=files,
            alignment=alignment,
            background_correction_matrices=background_correction_matrices,
            illumination_correction_matrices=illumination_correction_matrices,
        )

    def _assemble_tiles(self) -> list[Tile]:
        tiles = []
        for _i, row in self._files.iterrows():
            file = row["path"]
            time_point = row["t"] if "t" in row.index and row["t"] is not None else 0
            channel = row["channel"]
            metadata = load_metaseries_tiff_metadata(file)
            z = (
                1
                if self._z_spacing is None
                else row["z"] if row["z"] is not None else 1
            )

            bgcm = None
            if self._background_correction_matrices is not None:
                bgcm = self._background_correction_matrices[channel]

            icm = None
            if self._illumination_correction_matrices is not None:
                icm = self._illumination_correction_matrices[channel]

            tiles.append(
                Tile(
                    path=file,
                    shape=(metadata["pixel-size-y"], metadata["pixel-size-x"]),
                    position=TilePosition(
                        time=time_point,
                        channel=int(channel[1:]),
                        z=z,
                        y=int(
                            metadata["stage-position-y"]
                            / metadata["spatial-calibration-y"]
                        ),
                        x=int(
                            metadata["stage-position-x"]
                            / metadata["spatial-calibration-x"]
                        ),
                    ),
                    background_correction_matrix_path=bgcm,
                    illumination_correction_matrix_path=icm,
                )
            )
        return tiles

    def get_yx_spacing(self) -> tuple[float, float]:
        metadata = load_metaseries_tiff_metadata(self._files.iloc[0]["path"])
        return (metadata["spatial-calibration-y"], metadata["spatial-calibration-x"])

    def get_z_spacing(self) -> float | None:
        return self._z_spacing

    def get_axes(self) -> list[str]:
        axes = ["y", "x"]

        if "z" in self._files.columns:
            axes = ["z", *axes]

        if self._files["channel"].nunique() > 1:
            axes = ["c", *axes]

        if "t" in self._files.columns and self._files["t"].nunique() > 1:
            axes = ["t", *axes]

        return axes


class ImageXpressPlateAcquisition(PlateAcquisition):
    def __init__(
        self,
        acquisition_dir: Path | str,
        alignment: TileAlignmentOptions,
        background_correction_matrices: dict[str, Path | str] | None = None,
        illumination_correction_matrices: dict[str, Path | str] | None = None,
    ):
        super().__init__(
            acquisition_dir=acquisition_dir,
            alignment=alignment,
            background_correction_matrices=background_correction_matrices,
            illumination_correction_matrices=illumination_correction_matrices,
        )

    def _parse_files(self) -> pd.DataFrame:
        """Parse all files in the acquisition directory.

        Returns
        -------
        DataFrame
            Table of all files in the acquisition.
        """
        files = pd.DataFrame(
            ImageXpressPlateAcquisition._list_and_match_files(
                root_dir=self._acquisition_dir,
                root_re=self._get_root_re(),
                filename_re=self._get_filename_re(),
            )
        )

        # handle ImageXpress exported files:
        # MIPs are stored in ZStep_0 -> change 0 to None
        if "z" in files.columns:
            files.loc[files.z == "0", "z"] = None

        return files

    @staticmethod
    def _list_and_match_files(
        root_dir: Path | str,
        root_re: re.Pattern,
        filename_re: re.Pattern,
    ) -> list[list[dict[dict, str]]]:
        files = []
        for root, _, filenames in os.walk(root_dir):
            m_root = root_re.fullmatch(root)
            if m_root:
                for f in filenames:
                    m_filename = filename_re.fullmatch(f)
                    if m_filename:
                        row = m_root.groupdict()
                        row |= m_filename.groupdict()
                        if "channel" not in row or row["channel"] is None:
                            row["channel"] = "w1"
                        row["path"] = str(Path(root).joinpath(f))
                        files.append(row)
        return files

    @abstractmethod
    def _get_root_re(self) -> re.Pattern:
        """Regular expression for matching the root directory of the acquisition."""
        raise NotImplementedError

    @abstractmethod
    def _get_filename_re(self) -> re.Pattern:
        """Regular expression for matching the filename of the acquisition."""
        raise NotImplementedError

    def _build_well_acquisitions(self, files: pd.DataFrame) -> list[WellAcquisition]:
        return [
            ImageXpressWellAcquisition(
                files=files[files["well"] == well],
                alignment=self._alignment,
                z_spacing=self._get_z_spacing(),
                background_correction_matrices=self._background_correction_matrices,
                illumination_correction_matrices=self._illumination_correction_matrices,
            )
            for well in tqdm(files["well"].unique())
        ]

    @abstractmethod
    def _get_z_spacing(self) -> float | None:
        raise NotImplementedError

    def get_channel_metadata(self) -> dict[int, ChannelMetadata]:
        ch_metadata = {}
        _files = self._wells[0]._files
        for ch in _files["channel"].unique():
            channel_files = _files[_files["channel"] == ch]
            path = channel_files["path"].iloc[0]
            metadata = load_metaseries_tiff_metadata(path=path)
            index = int(ch[1:]) - 1
            if "Z Projection Method" in metadata:
                name = (
                    f"{metadata['Z Projection Method'].replace(' ', '-')}-Projection_"
                    f"{metadata['_IllumSetting_']}"
                )
            else:
                name = metadata["_IllumSetting_"]
            ch_metadata[index] = ChannelMetadata(
                channel_index=index,
                channel_name=name,
                display_color=rgb_to_hex(*wavelength_to_rgb(metadata["wavelength"])),
                spatial_calibration_x=metadata["spatial-calibration-x"],
                spatial_calibration_y=metadata["spatial-calibration-y"],
                spatial_calibration_units=metadata["spatial-calibration-units"],
                z_spacing=self._get_z_spacing(),
                wavelength=metadata["wavelength"],
                exposure_time=float(metadata["Exposure Time"].split(" ")[0]),
                exposure_time_unit=metadata["Exposure Time"].split(" ")[1],
                objective=metadata["_MagSetting_"],
            )

        assert min(ch_metadata.keys()) == 0, "Channel indices must start at 0."
        return ch_metadata


class SinglePlaneAcquisition(ImageXpressPlateAcquisition):
    """Parse top folder (single planes) of an acquisition of a MolecularDevices ImageXpress Micro Confocal system.

    Storage layout on disk for 2 wells with 2 fields and 2 channels:

    Option A (as stored by Microscope):

    MIP-2P-2sub --> {name} [Optional]
    └── 2022-07-05 --> {date} [Optional]
        └── 1075 --> {acquisition id}
            └── Timepoint_1 --> {t} [Optional]
                ├── MIP-2P-2sub_C05_s1_w146C9B2CD-0BB3-4B8A-9187-2805F4C90506.tif
                ├── MIP-2P-2sub_C05_s1_w1_thumb6EFE77C6-B96D-412A-9FD1-710DBDA32821.tif
                ├── MIP-2P-2sub_C05_s1_w2B90625C8-6EA7-4E54-8289-C539EB75263E.tif
                ├── MIP-2P-2sub_C05_s1_w2_thumbEDDF803A-AE5E-4190-8C06-F54341AEC4A6.tif
                ├── MIP-2P-2sub_C05_s2_w1E2913F7F-E229-4B6A-BFED-02BCF54561FA.tif
                ├── MIP-2P-2sub_C05_s2_w1_thumb72E3641A-C91B-4501-900A-245BAC58FF46.tif
                ├── MIP-2P-2sub_C05_s2_w241C38630-BCFD-4393-8706-58755CECE059.tif
                ├── MIP-2P-2sub_C05_s2_w2_thumb5377A5AC-9BBF-4BAF-99A2-24896E3373A2.tif
                ├── MIP-2P-2sub_C06_s1_w152C23B9A-EB4C-4DF6-8A7F-F4147A9E7DDE.tif
                ├── MIP-2P-2sub_C06_s1_w1_thumb541AA634-387C-4B84-B0D8-EE4CB1C88E81.tif
                ├── MIP-2P-2sub_C06_s1_w2FB0D7D9B-3EA0-445E-9A05-7D01154A9A5C.tif
                ├── MIP-2P-2sub_C06_s1_w2_thumb8FA1E466-57CD-4237-B09B-CAB48154647D.tif
                ├── MIP-2P-2sub_C06_s2_w1F365E60C-BCC2-4B74-9856-BCE07C8B0FD3.tif
                ├── MIP-2P-2sub_C06_s2_w1_thumb9652366E-36A0-4B7F-8B18-DA89D7DB41BD.tif
                ├── MIP-2P-2sub_C06_s2_w20EEC6AEA-1727-41E6-806C-40FF6AF68B6C.tif
                └── MIP-2P-2sub_C06_s2_w2_thumb710CD846-0185-4362-BBAF-C700AE0013B3.tif

    Image data is stored in {name}_{well}_{field}_w{channel}{md_id}.tif.
    The *_thumb*.tif files, used by Molecular Devices as preview, are ignored.

    Option B (as exported via software):

    test_Plate_3420 --> {name}_Plate_{acquisition id}
    └── Timepoint_1 --> {t} [Optional]
        ├── test_C05_s1_w1.TIF
        ├── test_C05_s1_w2.TIF
        ├── test_C05_s2_w1.TIF
        ├── test_C05_s2_w2.TIF
        ├── test_C06_s1_w1.TIF
        ├── test_C06_s1_w2.TIF
        ├── test_C06_s2_w1.TIF
        └── test_C06_s2_w2.TIF

    Image data is stored in {name}_{well}_{field}_w{channel}.TIF.
    """

    def __init__(
        self,
        acquisition_dir: Path | str,
        alignment: TileAlignmentOptions,
        background_correction_matrices: dict[str, Path | str] | None = None,
        illumination_correction_matrices: dict[str, Path | str] | None = None,
    ):
        super().__init__(
            acquisition_dir=acquisition_dir,
            alignment=alignment,
            background_correction_matrices=background_correction_matrices,
            illumination_correction_matrices=illumination_correction_matrices,
        )

    def _get_root_re(self) -> re.Pattern:
        return re.compile(
            r".*(?:[\/\\](?P<date>\d{4}-\d{2}-\d{2}))?[\/\\](?:(?P<plate_name>.*)_Plate_)?(?P<acq_id>\d+)(?:[\/\\]TimePoint_(?P<t>\d+))?(?:[\/\\]ZStep_0)?"
        )

    def _get_filename_re(self) -> re.Pattern:
        return re.compile(
            r"(?P<name>.*)_(?P<well>[A-Z]+\d{2})_?(?P<field>s\d+)?_?(?P<channel>w[1-9]{1})?(?!_thumb)(?P<md_id>[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12})?(?P<ext>\.(?i:tif))"
        )

    def _get_z_spacing(self) -> float | None:
        return None


class StackAcquisition(ImageXpressPlateAcquisition):
    """Image stack acquisition with a Molecular Devices ImageXpress Micro
    Confocal system.

    Storage layout on disk for 2 wells with 2 fields, 4 channels and 9 z-steps:

    OPTION A (as stored by Microscope):

    MIP-2P-2sub-Stack --> {name} [Optional]
    └── 2023-02-21 --> {date} [Optional]
        └── 1334 --> {acquisition id}
            └── Timepoint_1 --> {t} [Optional]
                ├── ZStep_1 --> {z}
                │   ├── Projection-Mix_E07_s1_w1E78EB128-BD0D-4D94-A6AD-3FF28BB1B105.tif
                │   ├── Projection-Mix_E07_s1_w1_thumb187DE64B-038A-4671-BF6B-683721723769.tif
                │   ├── Projection-Mix_E07_s1_w2C0A49256-E289-4C0F-ADC9-F7728ABDB141.tif
                │   ├── Projection-Mix_E07_s1_w2_thumb57D4B151-71BF-480E-8CC4-C23A2690B763.tif
                │   ├── Projection-Mix_E07_s1_w427CCB2E4-1BF4-45E7-8BC7-264B48EF9C4A.tif
                │   ├── Projection-Mix_E07_s1_w4_thumb555647D0-77F1-4A43-9472-AE509F95E236.tif
                │   ├── ...
                │   └── Projection-Mix_E08_s2_w4_thumbD2785594-4F49-464F-9F80-1B82E30A560A.tif
                ├── ...
                └── ZStep_9
                    ├── Projection-Mix_E07_s1_w1091EB8A5-272A-466D-B8A0-7547C6BA392B.tif
                    ├── ...
                    └── Projection-Mix_E08_s2_w2_thumb210C0D5D-C20E-484D-AFB2-EFE669A56B84.tif

    Image data is stored in {name}_{well}_{field}_w{channel}{md_id}.tif.
    The *_thumb*.tif files, used by Molecular Devices as preview, are ignored.

    OPTION B (as exported via software):

    test_Plate_3433 --> {name}_Plate_{acquisition id}
    └── Timepoint_1 --> {t} [Optional]
        ├── ZStep_0 (contains MIPs)
            └── ...
        ├── ZStep_1 --> {z}
        │   ├── test_E07_s1_w1.TIF
        │   ├── test_E07_s1_w2.TIF
        │   ├── test_E07_s1_w3.TIF
        │   ├── test_E07_s1_w4.TIF
        │   ├── test_E07_s2_w1.TIF
        │   ├── ...
        │   └── test_E08_s2_w4.TIF
        ├── ...
        └── ZStep_9
            ├── test_E07_s1_w1.TIF
            ├── ...
            └── test_E08_s2_w4.TIF

    Image data is stored in {name}_{well}_{field}_w{channel}.TIF.
    ZStep_0 contains MIPs and is ignored.
    """

    _z_spacing: float = None

    def __init__(
        self,
        acquisition_dir: Path | str,
        alignment: TileAlignmentOptions,
        background_correction_matrices: dict[str, Path | str] | None = None,
        illumination_correction_matrices: dict[str, Path | str] | None = None,
    ):
        super().__init__(
            acquisition_dir=acquisition_dir,
            alignment=alignment,
            background_correction_matrices=background_correction_matrices,
            illumination_correction_matrices=illumination_correction_matrices,
        )

    def _parse_files(self) -> pd.DataFrame:
        files = super()._parse_files()
        self._z_spacing = self._compute_z_spacing(files)
        return files

    def _get_root_re(self) -> re.Pattern:
        return re.compile(
            r".*(?:[\/\\](?P<date>\d{4}-\d{2}-\d{2}))?[\/\\](?:(?P<plate_name>.*)_Plate_)?(?P<acq_id>\d+)(?:[\/\\]TimePoint_(?P<t>\d+))?(?:[\/\\]ZStep_(?P<z>[1-9]\d*))"
        )

    def _get_filename_re(self) -> re.Pattern:
        return re.compile(
            r"(?P<name>.*)_(?P<well>[A-Z]+\d{2})_?(?P<field>s\d+)?_?(?P<channel>w[1-9]{1})?(?!_thumb)(?P<md_id>[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12})?(?P<ext>\.(?i:tif))"
        )

    def _get_z_spacing(self) -> float | None:
        return self._z_spacing

    def _compute_z_spacing(self, files: pd.DataFrame) -> float | None:
        assert "z" in files.columns, "No z column in files DataFrame."
        # check if files were exported via software
        if files.plate_name.iloc[0] is not None:
            # single planes and MIPs are duplicated to the entire stack
            # -> need to check in metadata, if the channel contains a stack or not
            for c in np.sort(files.channel.unique()):
                file = files[(files.channel == c) & (files.z == "2")].iloc[0]
                metadata = load_metaseries_tiff_metadata(file.path)
                # projections have no "Z Step" attribute
                if "Z Step" in metadata.keys():
                    # single-planes have metadata["Z Step"] == 1, for each duplicate
                    if metadata["Z Step"] == 2:
                        channel_with_stack = c
                        break
        else:
            channel_with_stack = np.sort(files[files["z"] == "2"]["channel"].unique())[
                0
            ]
        subset = files[files["channel"] == channel_with_stack]
        subset = subset[subset["well"] == np.sort(subset["well"].unique())[0]]
        first_field = np.sort(subset["field"].unique())[0]
        if first_field is not None:
            subset = subset[subset["field"] == np.sort(subset["field"].unique())[0]]

        plane_positions = []

        for _i, row in subset.iterrows():
            file = row["path"]
            if "z" in row and row["z"] is not None:
                metadata = load_metaseries_tiff_metadata(file)
                z_position = metadata["stage-position-z"]
                plane_positions.append(z_position)

        plane_positions = np.array(sorted(plane_positions), dtype=np.float32)

        precision = -Decimal(str(plane_positions[0])).as_tuple().exponent
        return np.round(np.mean(np.diff(plane_positions)), decimals=precision)


class MixedAcquisition(StackAcquisition):
    """Image stack acquisition with Projectsion acquired with a Molecular
    Devices ImageXpress Micro Confocal system.

    OPTION A (as stored by Microscope):

    MIP-2P-2sub-Stack --> {name} [Optional]
    └── 2023-02-21 --> {date}
        └── 1334 --> {acquisition id}
            ├── Projection-Mix_E07_s1_w1E94C24BD-45E4-450A-9919-257C714278F7.tif
            ├── Projection-Mix_E07_s1_w1_thumb4BFD4018-E675-475E-B5AB-2E959E6B6DA1.tif
            ├── ...
            ├── Projection-Mix_E08_s2_w3CCE83D85-0912-429E-9F18-716A085BB5BC.tif
            ├── Projection-Mix_E08_s2_w3_thumb4D88636E-181E-4AF6-BC53-E7A435959C8F.tif
            ├── ZStep_1
            │   ├── Projection-Mix_E07_s1_w1E78EB128-BD0D-4D94-A6AD-3FF28BB1B105.tif
            │   ├── Projection-Mix_E07_s1_w1_thumb187DE64B-038A-4671-BF6B-683721723769.tif
            │   ├── Projection-Mix_E07_s1_w2C0A49256-E289-4C0F-ADC9-F7728ABDB141.tif
            │   ├── Projection-Mix_E07_s1_w2_thumb57D4B151-71BF-480E-8CC4-C23A2690B763.tif
            │   ├── Projection-Mix_E07_s1_w427CCB2E4-1BF4-45E7-8BC7-264B48EF9C4A.tif
            │   ├── Projection-Mix_E07_s1_w4_thumb555647D0-77F1-4A43-9472-AE509F95E236.tif
            │   ├── ...
            │   └── Projection-Mix_E08_s2_w4_thumbD2785594-4F49-464F-9F80-1B82E30A560A.tif
            ├── ...
            └── ZStep_9
                ├── Projection-Mix_E07_s1_w1091EB8A5-272A-466D-B8A0-7547C6BA392B.tif
                ├── ...
                └── Projection-Mix_E08_s2_w2_thumb210C0D5D-C20E-484D-AFB2-EFE669A56B84.tif

    Image data is stored in {name}_{well}_{field}_w{channel}{md_id}.tif.
    The *_thumb*.tif files, used by Molecular Devices as preview, are ignored.

    OPTION B (as exported via software) will always export mixed acquisitions
    as stacks. -> Use StackAcquisition instead.
    """

    def __init__(
        self,
        acquisition_dir: Path | str,
        alignment: TileAlignmentOptions,
        background_correction_matrix: dict[str, NDArray] | None = None,
        illumination_correction_matrix: NDArray | None = None,
    ):
        super().__init__(
            acquisition_dir=acquisition_dir,
            alignment=alignment,
            background_correction_matrices=background_correction_matrix,
            illumination_correction_matrices=illumination_correction_matrix,
        )

    def _parse_files(self) -> pd.DataFrame:
        files = self._filter_mips(super()._parse_files())
        if files.plate_name.iloc[0] is not None:
            raise ValueError(
                "Data was exported via software. "
                "MixedAcquisition is not applicable in this case. "
                "Use StackAcquisition instead."
            )
        self._z_spacing = self._compute_z_spacing(files)
        return files

    def _filter_mips(self, files: pd.DataFrame) -> pd.DataFrame:
        """Remove MIP files if the whole stack was acquired."""
        _files = files.copy()
        for ch in _files["channel"].unique():
            channel_files = _files[_files["channel"] == ch]
            z_positions = channel_files["z"].unique()
            has_mip = None in z_positions
            has_stack = len(z_positions) > 1
            if has_mip and has_stack:
                _files.drop(
                    _files[(_files["channel"] == ch) & (_files["z"].isna())].index,
                    inplace=True,
                )
        return _files

    def _get_root_re(self) -> re.Pattern:
        return re.compile(
            r".*(?:[\/\\](?P<date>\d{4}-\d{2}-\d{2}))?[\/\\](?:(?P<plate_name>.*)_Plate_)?(?P<acq_id>\d+)(?:[\/\\]TimePoint_(?P<t>\d+))?(?:[\/\\]ZStep_(?P<z>[1-9]\d*))?.*"
        )

    def _get_filename_re(self) -> re.Pattern:
        return re.compile(
            r"(?P<name>.*)_(?P<well>[A-Z]+\d{2})_?(?P<field>s\d+)?_?(?P<channel>w[1-9]{1})?(?!_thumb)(?P<md_id>[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12})?(?P<ext>\.(?i:tif))"
        )



import numpy as np
from scipy.optimize import curve_fit
from skimage.measure import centroid


def ellipsoid_2D(coords, A, B, mu_x, mu_y, cxx, cxy, cyy):
    """
    Evaluates the given ellipsoid parameters at the provided coordinates.

    :param coords: at which the evaluation takes place
    :param A: amplitude
    :param B: offset
    :param mu_x: x location of the centroid
    :param mu_y: y location of the centroid
    :param cxx: covariance matrix term at 00
    :param cxy: covariance matrix term at 01
    :param cyy: covariance matrix term at 11
    :return:
    """
    inv = np.linalg.inv(np.array([[cxx, cxy], [cxy, cyy]]) + np.identity(2) * 1e-8)

    return (
        A
        * np.exp(
            -0.5
            * (
                inv[0, 0] * (coords[:, 1] - mu_x) ** 2
                + 2 * inv[0, 1] * (coords[:, 1] - mu_x) * (coords[:, 0] - mu_y)
                + inv[1, 1] * (coords[:, 0] - mu_y) ** 2
            )
        )
        + B
    )


def get_cov_matrix(img):
    """
    Compute covariance matrix of the provided image.

    :param img:
    :return: covariance matrix
    """

    def cov(x, y, i):
        return np.sum(x * y * i) / np.sum(i)

    y, x = np.meshgrid(
        np.arange(img.shape[0]),
        np.arange(img.shape[1]),
        indexing="ij",
    )
    cen = centroid(img)
    y = y.ravel() - cen[0]
    x = x.ravel() - cen[1]

    cxx = cov(x, x, img.ravel())
    cyy = cov(y, y, img.ravel())
    cxy = cov(x, y, img.ravel())

    return np.array([[cxx, cxy], [cxy, cyy]])


def get_estimates(data):
    """
    Estimate parameters of 2D Gaussian fit to the data.
    :param data:
    :return: (amplitude, offset, centroid-x, centroid-y, cxx, cxy, cyy)
    """
    max_ = data.max()
    mean_ = data.mean()
    cy, cx = centroid(data)
    cov = get_cov_matrix(data)
    return [
        max_ - mean_,
        mean_,
        cx,
        cy,
        cov[0, 0],
        cov[0, 1],
        cov[1, 1],
    ]


def get_coords(data):
    """
    Compute coordinate grid of provided 2D data.
    :param data:
    :return: coordinates
    """
    yy = np.arange(data.shape[0])
    xx = np.arange(data.shape[1])
    y, x = np.meshgrid(yy, xx, indexing="ij")
    return np.stack([y.ravel(), x.ravel()], -1)


def fit_gaussian_2d(data, coords):
    """
    Fit 2D ellipsoid to data.
    :param data: intensities
    :param coords: pixel coordinates
    :return: popt, pcov (see scipy.optimize.curve_fit)
    """
    return curve_fit(
        ellipsoid_2D,
        coords,
        data.ravel(),
        p0=get_estimates(data),
    )


def compute_fitted_matrix(coords, ellipsoid_parameters, shape):
    """
    Compute 2D ellipsoid for provided set of parameters
    :param coords: pixel coordinates
    :param ellipsoid_parameters:
    :param shape: of the final output
    :return: matrix
    """
    return ellipsoid_2D(coords, *ellipsoid_parameters).reshape(shape)



import numpy as np


def normalize_matrix(matrix):
    """
    Normalize matrix to [0, 1] by dividing it by its maximum value.

    :param matrix:
    :return: normalized matrix
    """
    return np.clip(matrix / matrix.max(), 0.0, 1.0)



from collections.abc import Callable
from copy import copy
from functools import partial

import numpy as np
from dask import array as da
from dask.array.core import normalize_chunks

from faim_ipa.stitching import BoundingBox5D, stitching_utils
from faim_ipa.stitching.tile import Tile


class DaskTileStitcher:
    """
    Stitch tiles into a single image using dask.
    """

    def __init__(
        self,
        tiles: list[Tile],
        chunk_shape: tuple[int, int] | tuple[int, int, int],
        output_shape: tuple[int, int, int, int, int] | None = None,
        dtype: np.dtype = np.uint16,
    ):
        """
        Parameters
        ----------
        tiles :
            Tiles to stitch.
        chunk_shape :
            Chunk shape in (Z)YX.
        output_shape :
            Shape of the output image. If None, the shape is computed from the tiles.
        dtype :
            Data type of the output image.
        """
        self.tiles: list[Tile] = stitching_utils.shift_to_origin(tiles)

        self.chunk_shape = (1,) * (5 - len(chunk_shape)) + chunk_shape
        self.dtype = dtype

        if output_shape is None:
            self._shape = self._compute_output_shape()
        else:
            self._shape = output_shape
        self._n_chunks = self._compute_number_of_chunks()
        self._block_to_tile_map = self._compute_block_to_tile_map()

    def _build_tiles_lut(self):
        lut = {}
        for tile in self.tiles:
            tcz_pos = (
                tile.position.time,
                tile.position.channel,
                tile.position.z,
            )
            if tcz_pos in lut:
                lut[tcz_pos].append(tile)
            else:
                lut[tcz_pos] = [tile]

        return lut

    def _compute_number_of_chunks(self):
        """
        Compute the number of chunks (blocks) in the stitched image.
        """
        return tuple(
            np.ceil(np.array(self._shape) / np.array(self.chunk_shape)).astype(int)
        )

    def _compute_block_to_tile_map(self):
        """
        Compute a map from block position to tiles that overlap with the block.
        """
        tiles_lut = self._build_tiles_lut()
        block_to_tile_map = {}
        for block_position in np.ndindex(self._n_chunks):
            block_to_tile_map[block_position] = []
            block_bbox = BoundingBox5D.from_pos_and_shape(
                position=tuple(block_position * np.array(self.chunk_shape)),
                shape=self.chunk_shape,
            )
            pos = (block_bbox.time_start, block_bbox.channel_start, block_bbox.z_start)
            if pos in tiles_lut:
                for tile in tiles_lut[pos]:
                    tile_bbox = BoundingBox5D.from_pos_and_shape(
                        position=tile.get_position(),
                        shape=(1,) * (5 - len(tile.shape)) + tile.shape,
                    )
                    if block_bbox.overlaps(tile_bbox):
                        block_to_tile_map[block_position].append(tile)

        return block_to_tile_map

    def _compute_output_shape(self):
        """
        Compute the shape of the stitched image.
        """
        tile_extents = [
            (tile.get_position() + np.array((1,) * (5 - len(tile.shape)) + tile.shape))
            for tile in self.tiles
        ]
        return tuple(np.max(tile_extents, axis=0))

    def get_stitched_dask_array(
        self,
        warp_func: Callable = stitching_utils.translate_tiles_2d,
        fuse_func: Callable = stitching_utils.fuse_mean,
        *,
        build_acquisition_mask: bool = False,
    ) -> da.array:
        """
        Build the dask array for the stitched image.

        Parameters
        ----------
        warp_func :
            Function which warps the tiles into the correct position.
        fuse_func :
            Function which fuses the warped tiles.

        Returns
        -------
            Dask array of the stitched image.
        """
        func = partial(
            stitching_utils.assemble_chunk,
            warp_func=warp_func,
            fuse_func=fuse_func,
            build_acquisition_mask=build_acquisition_mask,
        )

        return da.map_blocks(
            func=func,
            tile_map=copy(self._block_to_tile_map),
            dtype=copy(self.dtype),
            chunks=normalize_chunks(
                chunks=self.chunk_shape, shape=self._shape, dtype=self.dtype
            ),
        )

    def get_stitched_image(
        self,
        transform_func: Callable = stitching_utils.translate_tiles_2d,
        fuse_func: Callable = stitching_utils.fuse_mean,
        *,
        build_acquisition_mask: bool = False,
    ) -> np.ndarray:
        """
        Stitch the tiles into a single image.

        Note: This runs in 'synchronous' mode i.e. single threaded. If you
        want to use multiple threads, use get_stitched_dask_array() and
        compute the result using a configured dask scheduler.

        Parameters
        ----------
        transform_func :
            Function to transform the tiles into the stitched image.
        fuse_func :
            Function to fuse the transformed tiles.
        build_acquisition_mask :
            Whether to build an acquisition mask instead of the raw data.

        Returns
        -------
        Fused image.
        """
        stitched = self.get_stitched_dask_array(
            warp_func=transform_func,
            fuse_func=fuse_func,
            build_acquisition_mask=build_acquisition_mask,
        )
        return stitched.compute()



from abc import ABC, abstractmethod
from copy import copy

import numpy as np

from faim_ipa.stitching import stitching_utils
from faim_ipa.stitching.tile import Tile


class AbstractAlignment(ABC):
    _unaligned_tiles: list[Tile] = None
    _aligned_tiles: list[Tile] = None

    def __init__(self, tiles: list[Tile]) -> None:
        super().__init__()
        self._unaligned_tiles = stitching_utils.shift_to_origin(tiles)
        self._aligned_tiles = self._align(self._unaligned_tiles)

    @abstractmethod
    def _align(self, tiles: list[Tile]) -> list[Tile]:
        raise NotImplementedError

    def get_tiles(self) -> list[Tile]:
        return self._aligned_tiles


class StageAlignment(AbstractAlignment):
    """
    Align tiles using stage positions.
    """

    def _align(self, tiles: list[Tile]) -> list[Tile]:
        return tiles


class GridAlignment(AbstractAlignment):
    """
    Align tiles on a regular grid.
    """

    def _align(self, tiles: list[Tile]) -> list[Tile]:
        aligned_tiles = []

        tile_shape = tiles[0].shape

        grid_positions_y = set()
        grid_positions_x = set()
        tile_map = {}
        for tile in tiles:
            if tile.shape[-2:] != tile_shape[-2:]:
                message = f"All tiles must have the same YX shape. {tile.shape[-2:]} <=> {tile_shape[-2:]}"
                raise ValueError(message)
            y_pos = int(np.round(tile.position.y / tile_shape[-2]))
            x_pos = int(np.round(tile.position.x / tile_shape[-1]))
            if (y_pos, x_pos) in tile_map:
                tile_map[(y_pos, x_pos)].append(tile)
            else:
                tile_map[(y_pos, x_pos)] = [tile]
            grid_positions_y.add(y_pos)
            grid_positions_x.add(x_pos)

        grid_positions_y = sorted(grid_positions_y)
        grid_positions_x = sorted(grid_positions_x)
        for y_pos in grid_positions_y:
            for x_pos in grid_positions_x:
                if (y_pos, x_pos) in tile_map:
                    for unaligned_tile in tile_map[(y_pos, x_pos)]:
                        new_tile = copy(unaligned_tile)
                        new_tile.position.y = y_pos * tile_shape[-2]
                        new_tile.position.x = x_pos * tile_shape[-1]
                        aligned_tiles.append(new_tile)

        return aligned_tiles



from faim_ipa.hcs.cellvoyager.acquisition import (
    CellVoyagerWellAcquisition,
    StackAcquisition,
    ZAdjustedStackAcquisition,
)
from faim_ipa.hcs.cellvoyager.tile import StackedTile

__all__ = [
    "CellVoyagerWellAcquisition",
    "StackAcquisition",
    "ZAdjustedStackAcquisition",
    "StackedTile",
]



from skimage.measure.fit import BaseModel


class BoundingBox5D(BaseModel):
    """
    A 5D bounding box with integer coordinates.
        * time
        * channel
        * z
        * y
        * x
    """

    time_start: int
    time_end: int
    channel_start: int
    channel_end: int
    z_start: int
    z_end: int
    y_start: int
    y_end: int
    x_start: int
    x_end: int

    def __init__(
        self,
        time_start: int,
        time_end: int,
        channel_start: int,
        channel_end: int,
        z_start: int,
        z_end: int,
        y_start: int,
        y_end: int,
        x_start: int,
        x_end: int,
    ):
        super().__init__()
        self.time_start = time_start
        self.time_end = time_end
        self.channel_start = channel_start
        self.channel_end = channel_end
        self.z_start = z_start
        self.z_end = z_end
        self.y_start = y_start
        self.y_end = y_end
        self.x_start = x_start
        self.x_end = x_end

    @classmethod
    def from_pos_and_shape(
        cls,
        position: tuple[int, int, int, int, int],
        shape: tuple[int, int, int, int, int],
    ):
        """
        Create a bounding box from a position and shape.

        Parameters
        ----------
        position :
            Location of the bounding box.
        shape :
            Size of the bounding box

        Returns
        -------
        BoundingBox5D
        """
        return BoundingBox5D(
            time_start=position[0],
            time_end=position[0] + shape[0],
            channel_start=position[1],
            channel_end=position[1] + shape[1],
            z_start=position[2],
            z_end=position[2] + shape[2],
            y_start=position[3],
            y_end=position[3] + shape[3],
            x_start=position[4],
            x_end=position[4] + shape[4],
        )

    def overlaps_in_time(self, bbox: "BoundingBox5D") -> bool:
        return (self.time_start < bbox.time_end) and (bbox.time_start < self.time_end)

    def overlaps_in_channel(self, bbox: "BoundingBox5D") -> bool:
        return (self.channel_start < bbox.channel_end) and (
            bbox.channel_start < self.channel_end
        )

    def overlaps_in_z(self, bbox: "BoundingBox5D") -> bool:
        return (self.z_start < bbox.z_end) and (bbox.z_start < self.z_end)

    def overlaps_in_y(self, bbox: "BoundingBox5D") -> bool:
        return (self.y_start < bbox.y_end) and (bbox.y_start < self.y_end)

    def overlaps_in_x(self, bbox: "BoundingBox5D") -> bool:
        return (self.x_start < bbox.x_end) and (bbox.x_start < self.x_end)

    def overlaps(self, bbox: "BoundingBox5D") -> bool:
        """
        Check if this bounding box overlaps with another bounding box.

        Parameters
        ----------
        bbox :
            Bounding box to check.
        """
        if not self.overlaps_in_time(bbox):
            return False

        if not self.overlaps_in_channel(bbox):
            return False

        if not self.overlaps_in_z(bbox):
            return False

        if not self.overlaps_in_y(bbox):
            return False

        if not self.overlaps_in_x(bbox):
            return False

        return True



import os
import string
from copy import copy
from os.path import join

import numpy as np
import pandas as pd
import zarr
from mobie.metadata import (
    add_regions_to_dataset,
    add_source_to_dataset,
    add_view_to_dataset,
    get_image_display,
    get_merged_grid_source_transform,
    get_segmentation_display,
    read_dataset_metadata,
)
from mobie.tables import read_table
from skimage.measure import regionprops_table
from tqdm.auto import tqdm

from faim_ipa.histogram import UIntHistogram


def hex_to_rgba(h) -> str:
    """Convert a hex-color-code to RGBA values.

    :param h: A hex-color-code
    :return: RGBA color-code
    """
    if not isinstance(h, str):
        return "r=255,g=255,b=255,a=255"
    return f"r={int(h[0:2], 16)},g={int(h[2:4], 16)},b={int(h[4:6], 16)},a=255"


def to_position(well_name):
    r, c = well_name[0], well_name[1:]
    r = string.ascii_uppercase.index(r)
    c = int(c) - 1
    return [c, r]


def path_to_well(path: str):
    row, col = path.split("/")
    return f"{row}{col.zfill(2)}"


def add_wells_to_project(
    plate: zarr.Group,
    dataset_folder: str,
    well_group: str = "0",
    view_name: str = "default",
    label_suffix: str = "",
):
    """Add well groups of a ome-zarr plate to project

    :param plate: ome-zarr plate
    :param dataset_folder: mobie dataset folder
    :param well_group: zarr-index to zarr array. Defaults into zeroth field
    :param view_name: mobie view name
    :param label_suffix: suffix to distinguish empty views of the same well.
    :return:
    """

    sources = {}
    plate_hists = {}
    plate_colors = {}

    # Add wells as individual sources
    wells = []
    for well in tqdm(plate.attrs["plate"]["wells"]):
        attrs = plate[well["path"]][well_group].attrs.asdict()
        group_name = path_to_well(well["path"])
        wells.append(group_name)
        path = join(plate.store.path, well["path"], well_group)

        hists = [
            UIntHistogram.load(join(path, h_path)) for h_path in attrs["histograms"]
        ]

        for k, ch in enumerate(attrs["omero"]["channels"]):
            key = f"{ch['wavelength_id']}_{ch['label']}"
            if ch["label"] == "empty":
                key = key + label_suffix

            name = f"{group_name}_{key}"
            name = name.replace(" ", "_")

            add_source_to_dataset(
                dataset_folder=dataset_folder,
                source_type="image",
                source_name=name,
                image_metadata_path=path,
                file_format="ome.zarr",
                channel=k,
                view={},  # do not create default view for source
            )

            if key not in sources:
                sources[key] = [name]
            else:
                sources[key].append(name)

            if key not in plate_hists:
                plate_hists[key] = copy(hists[k])
            else:
                plate_hists[key].combine(hists[k])

            if key not in plate_colors:
                plate_colors[key] = hex_to_rgba(ch["color"])

    _add_well_regions(
        dataset_folder=dataset_folder,
        wells=wells,
    )

    _add_channel_plate_overviews(
        dataset_folder=dataset_folder,
        plate_hists=plate_hists,
        plate_colors=plate_colors,
        sources=sources,
        view_name=view_name,
    )


def add_labels_view(
    plate: zarr.Group,
    dataset_folder: str,
    *,
    well_group: str = "0",
    channel: int = 0,
    label_name: str = "default",
    view_name: str = "default_labels",
    extra_properties: tuple[str] = ("area",),
    add_empty_tables: bool = False,
):
    """Add merged grid segmentation view for labels of all wells in zarr

    :param plate: Zarr group representing an HCS plate
    :param dataset_folder: Dataset folder of the MoBIE project
    :param well_group: Path to subgroup within each well, e.g. '0' or '0/projections'
    :param channel: Channel in the well image to be added as segmentation view
    :param label_name: Name of the label subgroup in the Zarr file
    :param view_name: View of the MoBIE dataset, will be updated in place
    :param extra_properties: Property names to be added to regionprops measurement table
    :param add_empty_tables: Write tables for empty segmentations
    """
    # add sources for each label image
    sources = []
    n_objects = []
    for well in tqdm(plate.attrs["plate"]["wells"]):
        path = join(plate.store.path, well["path"], well_group, "labels", label_name)
        group_name = path_to_well(well["path"])
        name = f"{group_name}_{label_name}"
        name = name.replace(" ", "_")

        # measure regionprops
        label_img = plate[well["path"]][well_group]["labels"][label_name][0][channel]
        datasets = plate[well["path"]][well_group]["labels"][label_name].attrs.asdict()[
            "multiscales"
        ][0]["datasets"]
        spacing = datasets[0]["coordinateTransformations"][0]["scale"]
        props = regionprops_table(
            label_img[np.newaxis, :],
            properties=("label", "centroid", *extra_properties),
            spacing=spacing,
        )
        if not add_empty_tables and len(props["label"]) == 0:
            continue

        # write default.tsv to dataset_folder/tables/name
        # TODO: reconcile once saving table data inside zarr is possible
        table_folder = join(dataset_folder, "tables", name)
        os.makedirs(table_folder, exist_ok=True)

        table_path = join(table_folder, "default.tsv")
        table = pd.DataFrame(props)
        table.to_csv(table_path, sep="\t", index=False)
        n_objects.append(len(table.index))

        add_source_to_dataset(
            dataset_folder=dataset_folder,
            source_type="segmentation",
            source_name=name,
            image_metadata_path=path,
            file_format="ome.zarr",
            channel=channel,
            table_folder=table_folder,
            view={},  # do not create default view for source
            suppress_warnings=True,
        )
        sources.append(name)

    # get view 'view_name' from dataset
    dataset_metadata = read_dataset_metadata(dataset_folder=dataset_folder)
    view = dataset_metadata["views"][view_name]

    # get_merged_grid_source_transform for list of sources
    view["sourceTransforms"].append(
        get_merged_grid_source_transform(
            sources=sources,
            merged_source_name=f"merged_segmentation_view_{view_name}_{label_name}",
            positions=[to_position(src[:3]) for src in sources],
        )
    )

    # get_segmentation_display for grid source and append to sourceDisplays
    view["sourceDisplays"].append(
        get_segmentation_display(
            name=label_name,
            sources=[f"merged_segmentation_view_{view_name}_{label_name}"],
            visible=False,
        )
    )

    # update view in original dataset
    add_view_to_dataset(
        dataset_folder=dataset_folder,
        view_name=view_name,
        view=view,
    )

    # update wells table
    wells_table_path = join(dataset_folder, "tables", "wells", "default.tsv")
    wells_table = read_table(wells_table_path)
    wells_table[f"n_objects_{label_name}"] = n_objects
    wells_table.to_csv(wells_table_path, sep="\t", index=False)


def compute_aggregate_table_values(
    dataset_folder,
    table_suffix,
    aggregation_dict=None,
):
    """Aggregate all tables with given suffix, and write summarized values to wells table.

    :param dataset_folder: location of the MoBIE dataset to be modified
    :param table_suffix: common suffix of all tables to be included (e.g. 'my_seg' for 'A01_my_seg' etc.)
    :param aggregation_dict: mapping of column names to (one or multiple) aggregation methods
    """
    if aggregation_dict is None:
        aggregation_dict = {"area": ["mean", "min", "max"]}

    # read wells table and get list of wells (region_id)
    wells_table_path = join(dataset_folder, "tables", "wells", "default.tsv")
    wells_table = read_table(wells_table_path)
    wells = wells_table["region_id"]

    # read and concatenate each table from segmentations
    tables = []
    for well in wells:
        well_table = read_table(
            join(dataset_folder, "tables", f"{well}_{table_suffix}", "default.tsv")
        )
        well_table["region_id"] = well
        tables.append(well_table)
    joined_table = pd.concat(tables)

    # aggregate given columns with given functions
    summary = joined_table.groupby("region_id").aggregate(aggregation_dict)

    # flatten table index
    summary.columns = ["_".join(headers) for headers in summary.columns.to_flat_index()]
    # add suffix to column names
    summary.columns = [f"{header}_{table_suffix}" for header in summary.columns]

    # join with original wells table
    wells_table.join(summary, on="region_id").to_csv(
        wells_table_path, sep="\t", index=False
    )


def _add_channel_plate_overviews(
    dataset_folder, plate_hists, plate_colors, sources, view_name
):
    default = {
        "isExclusive": True,
        "sourceDisplays": [],
        "sourceTransforms": [],
        "uiSelectionGroup": "bookmarks",
    }
    for i, ch in enumerate(sources.keys()):
        name = ch.replace(" ", "_")
        default["sourceDisplays"].append(
            get_image_display(
                name=f"Plate_{name}",
                sources=[f"merged_view_plate_{name}"],
                color=plate_colors[ch],
                contrastLimits=[
                    plate_hists[ch].quantile(0.01),
                    plate_hists[ch].quantile(0.99),
                ],
                visible=i == 0,
            )
        )
        default["sourceTransforms"].append(
            get_merged_grid_source_transform(
                sources=list(sources[ch]),
                merged_source_name=f"merged_view_plate_{name}",
                positions=[to_position(src[:3]) for src in sources[ch]],
            )
        )

    default["sourceDisplays"].append(
        {
            "regionDisplay": {
                "opacity": 0.5,
                "lut": "glasbey",
                "name": "Wells",
                "sources": _get_well_sources_per_channel(sources),
                "tableSource": "wells",
                "visible": True,
                "showAsBoundaries": False,
            }
        }
    )

    add_view_to_dataset(
        dataset_folder=dataset_folder,
        view_name=view_name,
        view=default,
    )


def _get_well_sources_per_channel(sources):
    wells_per_channel = {}
    for ch in sources:
        for well in sources[ch]:
            if well[:3] not in wells_per_channel:
                wells_per_channel[well[:3]] = [well]
            else:
                wells_per_channel[well[:3]].append(well)

    return wells_per_channel


def _add_well_regions(dataset_folder, wells):
    well_table = pd.DataFrame(
        {
            "region_id": wells,
            "treatment": [
                "Unknown",
            ]
            * len(wells),
        }
    )
    add_regions_to_dataset(dataset_folder, "wells", well_table)



from pathlib import Path

from numpy._typing import ArrayLike
from tifffile import tifffile


def load_metaseries_tiff_metadata(path: Path) -> dict:
    """Load parts of the metadata of a metaseries tiff file.

    The following metadata is collected:
    * pixel-size-x
    * pixel-size-y
    * _IllumSetting_
    * spatial-calibration-x
    * spatial-calibration-y
    * spatial-calibration-units
    * ImageXpress Micro X
    * ImageXpress Micro Y
    * ImageXpress Micro Z
    * z-position
    * _MagNA_
    * _MagSetting_
    * Exposure Time
    * Lumencor Cyan Intensity
    * Lumencor Green Intensity
    * Lumencor Red Intensity
    * Lumencor Violet Intensity
    * Lumencor Yellow Intensity
    * ShadingCorrection
    * stage-label
    * SiteX
    * SiteY
    * wavelength
    * Z Step (if existent)
    * Z Projection Method (if existent)

    :param path:
    :return:
    image_data, metadata-dict
    """
    with tifffile.TiffFile(path) as tiff:
        assert tiff.is_metaseries, f"{path} is not a metamorph file."
        selected_keys = [
            "pixel-size-x",
            "pixel-size-y",
            "_IllumSetting_",
            "spatial-calibration-x",
            "spatial-calibration-y",
            "spatial-calibration-units",
            "ImageXpress Micro X",
            "ImageXpress Micro Y",
            "ImageXpress Micro Z",
            "_MagNA_",
            "_MagSetting_",
            "Exposure Time",
            "ShadingCorrection",
            "stage-label",
            "SiteX",
            "SiteY",
            "wavelength",
            "Z Step",  # optional
            "Z Projection Method",  # optional
            "Z Projection Step Size",  # optional
        ]
        plane_info = tiff.metaseries_metadata["PlaneInfo"]
        metadata = {
            k: plane_info[k]
            for k in plane_info
            if k in selected_keys or k.endswith("Intensity")
        }
        metadata["stage-position-x"] = metadata["ImageXpress Micro X"]
        metadata["stage-position-y"] = metadata["ImageXpress Micro Y"]
        metadata["stage-position-z"] = metadata["ImageXpress Micro Z"]

        metadata.pop("ImageXpress Micro X")
        metadata.pop("ImageXpress Micro Y")
        metadata.pop("ImageXpress Micro Z")

    return metadata


def load_metaseries_tiff(path: Path) -> tuple[ArrayLike, dict]:
    with tifffile.TiffFile(path) as tiff:
        data = tiff.asarray()
    metadata = load_metaseries_tiff_metadata(path=path)
    metadata["PixelType"] = str(data.dtype)
    return data, metadata



from decimal import Decimal
from pathlib import Path
from typing import Any

import numpy as np
from defusedxml import ElementTree
from defusedxml.ElementTree import parse

from faim_ipa.io.metadata import ChannelMetadata
from faim_ipa.utils import rgb_to_hex, wavelength_to_rgb

SCHEMA = "{http://www.openmicroscopy.org/Schemas/OME/2016-06}"


def get_z_spacing(metadata: ElementTree) -> float | None:
    """
    Get the Z spacing from the first image in the XML metadata.

    Parameters
    ----------
    metadata :
        XML metadata

    Returns
    -------
        Z spacing or None if there is only one Z plane
    """
    image_0 = next(metadata.iter(f"{SCHEMA}Image"))
    pixels = next(image_0.iter(f"{SCHEMA}Pixels"))
    z_positions = {}
    for plane in pixels.iter(f"{SCHEMA}Plane"):
        z_positions[plane.get("TheZ")] = float(plane.get("PositionZ"))

    if len(z_positions) == 1:
        return None
    z_positions = list(z_positions.values())
    precision = -Decimal(str(z_positions[0])).as_tuple().exponent
    z_spacing = np.round(np.diff(z_positions).mean(), precision)
    return z_spacing


def get_yx_spacing(metadata: ElementTree) -> tuple[float, float]:
    """
    Get the YX spacing from the first image in the XML metadata.

    Parameters
    ----------
    metadata :
        XML metadata

    Returns
    -------
        YX spacing
    """
    image_0 = next(metadata.iter(f"{SCHEMA}Image"))
    pixels = next(image_0.iter(f"{SCHEMA}Pixels"))
    return (
        float(pixels.get("PhysicalSizeY")),
        float(pixels.get("PhysicalSizeX")),
    )


def get_exposure_time(metadata: ElementTree) -> tuple[float, str]:
    """
    Get the exposure time and unit from the first image in the XML metadata.

    Parameters
    ----------
    metadata :
        XML metadata

    Returns
    -------
        Exposure time and unit
    """
    image_0 = next(metadata.iter(f"{SCHEMA}Image"))
    pixels = next(image_0.iter(f"{SCHEMA}Pixels"))
    plane_0 = next(pixels.iter(f"{SCHEMA}Plane"))
    return plane_0.get("ExposureTime"), plane_0.get("ExposureTimeUnit")


def get_channels(metadata: ElementTree) -> dict[str, ChannelMetadata]:
    """
    Get the channel metadata from the XML metadata.

    Parameters
    ----------
    metadata :
        XML metadata

    Returns
    -------
        Channel metadata for each channel.
    """
    image_0 = next(metadata.iter(f"{SCHEMA}Image"))
    pixels = next(image_0.iter(f"{SCHEMA}Pixels"))
    channels = [channel.attrib for channel in pixels.iter(f"{SCHEMA}Channel")]
    objective = next(image_0.iter(f"{SCHEMA}ObjectiveSettings")).get("ID")
    yx_spacing = get_yx_spacing(metadata)
    exposure_time, exposure_time_unit = get_exposure_time(metadata)

    ch_metadata = {}
    for channel in channels:
        idx = int(channel["ID"].split(":")[-1])
        if "EmissionWavelength" in channel:
            wavelength = int(channel["EmissionWavelength"])
            display_color = rgb_to_hex(*wavelength_to_rgb(wavelength))
        else:
            wavelength = None
            display_color = "#ffffff"
        ch_metadata[f"w{idx+1}"] = ChannelMetadata(
            channel_index=idx,
            channel_name=channel["Name"],
            display_color=display_color,
            spatial_calibration_x=yx_spacing[1],
            spatial_calibration_y=yx_spacing[0],
            spatial_calibration_units="um",
            z_spacing=get_z_spacing(metadata),
            wavelength=wavelength,
            exposure_time=exposure_time,
            exposure_time_unit=exposure_time_unit,
            objective=objective,
        )

    return ch_metadata


def get_stage_positions(
    metadata: ElementTree,
) -> dict[str, tuple[float, float]]:
    """
    Get the stage positions for each image from the XML metadata.

    Parameters
    ----------
    metadata :
        XML metadata

    Returns
    -------
        Stage positions for each image.
    """
    positions = {}
    for i, image in enumerate(metadata.iter(f"{SCHEMA}Image")):
        image_id = image.get("ID")
        index = int(image_id.split(":")[-1])
        assert index == i, f"Expected index {i} but got {index}"

        try:
            stage_label = next(image.iter(f"{SCHEMA}StageLabel")).attrib

            y_pos = float(stage_label["Y"])
            x_pos = float(stage_label["X"])
            positions[str(i + 1)] = (y_pos, x_pos)
        except (StopIteration, KeyError):
            # No stage positions in metadata
            positions[str(i + 1)] = (0, 0)

    return positions


def parse_basic_metadata(companion_file: Path | str) -> dict[str, Any]:
    """
    Parse the basic metadata from the XML companion file.

    Note: Only the first entries are parsed, and it is assumed that the
    extracted metadata is the same for all images.
    If you want to parse the complete metadata, use ome-types instead.

    ome-types: https://ome-types.readthedocs.io/en/latest/usage/

    Parameters
    ----------
    companion_file :
        Path to the XML companion file.

    Returns
    -------
        Basic metadata:
        - z_spacing
        - yx_spacing
        - channels
        - stage_positions
    """
    with open(companion_file, "rb") as f:
        root = parse(f).getroot()
        try:
            z_spacing = get_z_spacing(root)
        except StopIteration:
            z_spacing = None

        return {
            "z_spacing": z_spacing,
            "yx_spacing": get_yx_spacing(root),
            "channels": get_channels(root),
            "stage_positions": get_stage_positions(root),
        }



import logging
import os
from datetime import datetime
from pathlib import Path
from typing import TypeVar

import questionary
import yaml
from pydantic import (
    BaseModel,
    TypeAdapter,
    ValidationError,
    field_serializer,
    field_validator,
)
from questionary import ValidationError as QuestionaryValidationError
from questionary import Validator


def wavelength_to_rgb(wavelength, gamma=0.8):
    """This converts a given wavelength of light to an
    approximate RGB color value. The wavelength must be given
    in nanometers in the range from 380 nm through 750 nm
    (789 THz through 400 THz).
    Based on code by Dan Bruton
    http://www.physics.sfasu.edu/astro/color/spectra.html

    Obtained from https://gist.github.com/error454/65d7f392e1acd4a782fc
    """

    wavelength = float(wavelength)
    if 380 <= wavelength <= 440:  # noqa: PLR2004
        attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
        r = ((-(wavelength - 440) / (440 - 380)) * attenuation) ** gamma
        g = 0.0
        b = (1.0 * attenuation) ** gamma
    elif 440 <= wavelength <= 490:  # noqa: PLR2004
        r = 0.0
        g = ((wavelength - 440) / (490 - 440)) ** gamma
        b = 1.0
    elif 490 <= wavelength <= 510:  # noqa: PLR2004
        r = 0.0
        g = 1.0
        b = (-(wavelength - 510) / (510 - 490)) ** gamma
    elif 510 <= wavelength <= 580:  # noqa: PLR2004
        r = ((wavelength - 510) / (580 - 510)) ** gamma
        g = 1.0
        b = 0.0
    elif 580 <= wavelength <= 645:  # noqa: PLR2004
        r = 1.0
        g = (-(wavelength - 645) / (645 - 580)) ** gamma
        b = 0.0
    elif 645 <= wavelength <= 750:  # noqa: PLR2004
        attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
        r = (1.0 * attenuation) ** gamma
        g = 0.0
        b = 0.0
    else:
        r = 0.0
        g = 0.0
        b = 0.0
    r *= 255
    g *= 255
    b *= 255
    return int(r), int(g), int(b)


def rgb_to_hex(r, g, b):
    """Convert RGB values into hex number."""
    return f"{r:02x}{g:02x}{b:02x}"


def create_logger(name: str, *, include_timestamp: bool = True) -> logging.Logger:
    """
    Create logger which logs to `<timestamp>-<name>.log` inside the current
    working directory.

    Parameters
    ----------
    name
        Name of the logger instance.
    include_timestamp
        Whether to include the timestamp in the log file name.

    Returns
    -------
    logging.Logger
        Logger instance
    """
    logger = logging.getLogger(name=name)
    now = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    handler = logging.FileHandler(
        f"{now}-{name}.log" if include_timestamp else f"{name}.log"
    )
    formatter = logging.Formatter(
        fmt="{asctime} - {name} - {levelname} - {message}",
        style="{",
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    return logger


def get_git_root() -> Path:
    """
    Recursively search for the directory containing the `.git` folder.

    Returns
    -------
    Path
        Path to the root of the git repository. None if not found.
    """
    parent_dir = Path.cwd()
    while not (parent_dir / ".git").exists():
        if parent_dir == parent_dir.parent:
            return None  # reached root directory
        parent_dir = parent_dir.parent

    return parent_dir


def resolve(relative_path: Path, reference: Path) -> Path:
    """
    Takes a relative path and resolves it relative to the reference directory.

    Parameters
    ----------
    relative_path
        Path relative to the reference.
    reference
        Reference path.

    Returns
    -------
    Path
        Absolute path to the file.
    """
    if relative_path.is_absolute():
        return relative_path
    return (reference / relative_path).absolute()


def make_relative(path: Path, reference: Path):
    """
    Convert an absolute path to a path relative to the reference directory.

    Parameters
    ----------
    path
        Absolute path to a file.
    reference
        Path to the reference directory.

    Returns
    -------
    pathlib.Path
        Path relative to the reference.
    """
    try:
        # requires Python >= 3.12
        return path.relative_to(reference, walk_up=True)
    except (ValueError, TypeError):
        # fallback for Python < 3.12
        return Path(os.path.relpath(path, reference))


def prompt_with_questionary(
    model: "IPAConfig",
    defaults: dict | None = None,
):
    schema = model.model_json_schema()
    defaults = defaults or {}
    responses = {}

    for field_name, field_info in schema["properties"].items():
        description = field_info.get("description", field_name)
        field_type = field_info["type"]
        default_value = defaults.get(field_name, "")

        if field_type == "string":
            if field_info.get("format") == "path":
                default_path = str(resolve(Path(default_value), model.reference_dir()))
                responses[field_name] = Path(
                    questionary.path(
                        f"Enter {description} [{default_value}]",
                        validate=PathValidator(
                            model=model,
                            field_name=field_name,
                        ),
                        default=default_path,
                    ).ask()
                ).absolute()
            elif field_info.get("format") == "directory-path":
                default_path = str(resolve(Path(default_value), model.reference_dir()))
                responses[field_name] = Path(
                    questionary.path(
                        f"Enter {description} (directory) [{default_value}]",
                        validate=PathValidator(
                            model=model,
                            field_name=field_name,
                        ),
                        default=default_path,
                    ).ask()
                ).absolute()
            else:
                responses[field_name] = questionary.text(
                    f"Enter {description} [{default_value}]",
                    validate=QuestionaryPydanticValidator(
                        model=model, field_name=field_name
                    ),
                    default=default_value,
                ).ask()
        elif field_type == "integer":
            min_val = field_info.get("minimum", None)
            max_val = field_info.get("maximum", None)
            prompt_message = f"{description} ({f'minimum: {min_val}' if min_val else ''}-{f'maximum: {max_val}' if max_val else ''}) [{default_value}]"
            responses[field_name] = int(
                questionary.text(
                    prompt_message,
                    validate=QuestionaryPydanticValidator(
                        model=model, field_name=field_name
                    ),
                    default=str(default_value),
                ).ask()
            )
        elif field_type == "number":
            min_val = field_info.get("minimum", None)
            max_val = field_info.get("maximum", None)
            prompt_message = f"{description} ({f'minimum: {min_val}' if min_val else ''}-{f'maximum: {max_val}' if max_val else ''}) [{default_value}]"
            responses[field_name] = float(
                questionary.text(
                    prompt_message,
                    validate=QuestionaryPydanticValidator(
                        model=model, field_name=field_name
                    ),
                    default=str(default_value),
                ).ask()
            )
        elif field_type == "boolean":
            responses[field_name] = questionary.confirm(
                prompt_message,
                default=default_value,
            ).ask()
        else:
            msg = f"Unknown field type: {field_type}"
            raise ValueError(msg)

    return model(**responses)


class QuestionaryPydanticValidator(Validator):
    def __init__(self, model: BaseModel, field_name: str):
        self.field_name = field_name
        self.model = model
        self.field_info = model.model_fields[field_name]
        self.type_adapter = TypeAdapter(self.field_info.annotation)

    def _preprocess(self, value):
        return value

    def validate(self, document):
        try:
            value = self._preprocess(self.type_adapter.validate_python(document.text))
            self.model.__pydantic_validator__.validate_assignment(
                self.model.model_construct(), self.field_name, value
            )
        except ValidationError as e:
            raise QuestionaryValidationError(
                message=f"Invalid value for field: {e.errors()[0]['msg']}"
            ) from e


class PathValidator(QuestionaryPydanticValidator):
    def __init__(self, model: BaseModel, field_name: str):
        super().__init__(model, field_name)

    def _preprocess(self, value):
        return Path(value).absolute()


T = TypeVar("T", bound="IPAConfig")


class IPAConfig(BaseModel):

    @field_serializer("*")
    @classmethod
    def path_relative_to_git(cls, value):
        if isinstance(value, Path):
            try:
                return str(make_relative(value, cls.reference_dir()))
            except ValueError:
                return str(value)
        return value

    @field_validator("*", mode="before")
    @classmethod
    def git_relative_path_to_absolute(cls, value, info):
        field_name = info.field_name
        field_type = cls.__annotations__[field_name]
        if isinstance(field_type, type) and issubclass(field_type, Path):
            return resolve(Path(value), cls.reference_dir())
        if hasattr(field_type, "__metadata__") and issubclass(
            field_type.__origin__, Path
        ):
            return resolve(Path(value), cls.reference_dir())
        return value

    @staticmethod
    def reference_dir():
        return get_git_root()

    @staticmethod
    def config_name():
        return "config.yml"

    def save(self, config_file=None):
        config_file = config_file or Path.cwd() / self.config_name()
        with open(config_file, "w") as f:
            yaml.safe_dump(self.model_dump(), f, sort_keys=False)

    @classmethod
    def load(cls: type[T], config_file=None) -> T:
        config_file = config_file or Path.cwd() / cls.config_name()
        with open(config_file) as f:
            return cls(**yaml.safe_load(f))



from pathlib import Path

import numpy as np
from numpy._typing import NDArray
from tifffile import imread

from faim_ipa.stitching.tile import Tile, TilePosition


class StackedTile(Tile):
    def __init__(
        self,
        paths: list[Path | str],
        shape: tuple[int, int, int],
        dtype: np.dtype,
        position: TilePosition,
        background_correction_matrix_path: Path | str | None = None,
        illumination_correction_matrix_path: Path | str | None = None,
    ):
        super().__init__(
            path=None,
            shape=(len(paths),) + shape[1:],
            position=position,
            background_correction_matrix_path=background_correction_matrix_path,
            illumination_correction_matrix_path=illumination_correction_matrix_path,
        )
        self._paths = paths
        self._dtype = dtype

    def load_data(self):
        data = np.zeros(self.shape, dtype=self._dtype)
        for i, path in enumerate(self._paths):
            if path:
                plane = imread(path)
                plane = self._apply_background_correction(plane)
                plane = self._apply_illumination_correction(plane)
                data[i] = plane

        return data

    def load_data_mask(self) -> NDArray:
        data = np.zeros(self.shape, dtype=bool)
        for i, path in enumerate(self._paths):
            if path:
                data[i] = True

        return data



import json
import shutil
import subprocess
import sys
from inspect import cleandoc

GB = 1024 * 1024 * 1024


def min_cache_size_gb(gb: int = 2):
    info = subprocess.run(
        ["pixi", "info", "--json"],  # noqa: S607
        capture_output=True,
        text=True,
        check=False,
    )
    cache_dir = json.loads(info.stdout)["cache_dir"]

    if shutil.disk_usage(cache_dir).free < gb * GB:
        sys.tracebacklimit = 0
        message = cleandoc(
            f"""
        Disk space in cache directory is below {gb} GB.

        PIXI_CACHE_DIR: {cache_dir}

        Did you initialize your session correctly ('source ./init.sh') ?
        """
        )
        raise RuntimeError(message)
    sys.exit(0)


if __name__ == "__main__":
    if len(sys.argv) > 1:
        min_cache_size_gb(gb=int(sys.argv[1]))
    min_cache_size_gb()



from __future__ import annotations

import os
from os.path import join
from pathlib import Path
from typing import TYPE_CHECKING

import dask.array as da
import zarr
from dask.distributed import Client, wait
from numcodecs import Blosc
from ome_zarr.format import CurrentFormat
from ome_zarr.io import parse_url
from ome_zarr.writer import (
    Axes,
    write_multiscales_metadata,
    write_plate_metadata,
    write_well_metadata,
)
from pint import Unit
from pydantic import BaseModel

from faim_ipa import dask_utils
from faim_ipa.hcs.plate import PlateLayout, get_rows_and_columns
from faim_ipa.stitching import stitching_utils

if TYPE_CHECKING:
    from collections.abc import Callable

    from faim_ipa.hcs.acquisition import PlateAcquisition


class NGFFPlate(BaseModel):
    root_dir: Path | str
    name: str
    layout: PlateLayout
    order_name: str
    barcode: str


class ConvertToNGFFPlate:
    """
    Convert a plate acquisition to an NGFF plate.
    """

    _ngff_plate: NGFFPlate

    def __init__(
        self,
        ngff_plate: NGFFPlate,
        yx_binning: int = 1,
        warp_func: Callable = stitching_utils.translate_tiles_2d,
        fuse_func: Callable = stitching_utils.fuse_mean,
        client: Client = None,
    ):
        """
        Parameters
        ----------
        ngff_plate :
            NGFF plate information.
        yx_binning :
            YX binning factor.
        warp_func :
            Function used to warp tile images.
        fuse_func :
            Function used to fuse tile images.
        client :
            Dask client used for the conversion.
        """
        assert (
            isinstance(yx_binning, int) and yx_binning >= 1
        ), "yx_binning must be an integer >= 1."
        self._ngff_plate = ngff_plate
        self._yx_binning = yx_binning
        self._warp_func = warp_func
        self._fuse_func = fuse_func
        self._client = client

    def create_zarr_plate(
        self, plate_acquisition: PlateAcquisition, wells: list[str] | None = None
    ) -> zarr.Group:
        """
        Create empty NGFF zarr plate.

        Note: Loads the plate from disk if it already exists.

        Parameters
        ----------
        plate_acquisition :
            A single plate acquisition.
        wells :
            List of wells to build. If None, all wells are built.
        """
        plate_path = join(self._ngff_plate.root_dir, self._ngff_plate.name + ".zarr")
        if not os.path.exists(plate_path):
            os.makedirs(plate_path, exist_ok=False)
            store = parse_url(plate_path, mode="w").store
            plate = zarr.group(store=store)

            rows, cols = get_rows_and_columns(layout=self._ngff_plate.layout)

            write_plate_metadata(
                plate,
                columns=cols,
                rows=rows,
                wells=[
                    f"{w[0]}/{w[1:]}" for w in plate_acquisition.get_well_names(wells)
                ],
                name=self._ngff_plate.name,
                field_count=1,
            )

            attrs = plate.attrs.asdict()
            attrs["order_name"] = self._ngff_plate.order_name
            attrs["barcode"] = self._ngff_plate.barcode
            plate.attrs.put(attrs)
            return plate
        store = parse_url(plate_path, mode="w").store
        return zarr.group(store=store)

    def run(
        self,
        plate: zarr.Group,
        plate_acquisition: PlateAcquisition,
        wells: list[str] | None = None,
        well_sub_group: str = "0",
        chunks: tuple[int, int] | tuple[int, int, int] = (2048, 2048),
        max_layer: int = 3,
        storage_options: dict | None = None,
        *,
        build_acquisition_mask: bool = False,
        overwrite: bool = False,
    ):
        """
        Convert a plate acquisition to an NGFF plate.

        Parameters
        ----------
        plate_acquisition :
            A single plate acquisition.
        well_sub_group :
            Name of the well subgroup.
        chunks :
            Chunk size in (Z)YX.
        max_layer :
            Maximum layer of the resolution pyramid layers.
        storage_options :
            Zarr storage options.
        build_acquisition_mask :
            Writes a boolean mask instead of the image data, indicating where the image data is present.
        overwrite :
            Overwrite existing data.

        Returns
        -------
            zarr.Group of the plate.
        """
        assert 2 <= len(chunks) <= 3, "Chunks must be 2D or 3D."
        assert len(chunks) == len(
            plate_acquisition.get_well_acquisitions()[0].get_tiles()[0].shape
        ), "Chunks must have the same number of dimensions as the tile shape."
        well_acquisitions = plate_acquisition.get_well_acquisitions(wells)

        for well_acquisition in well_acquisitions:
            well_group = self._create_well_group(
                plate,
                well_acquisition,
                well_sub_group,
                add_to_well_images=not build_acquisition_mask,
            )
            group = well_group[well_sub_group]
            self._write_stitched_image(
                group,
                chunks,
                plate_acquisition,
                storage_options,
                well_acquisition,
                build_acquisition_mask=build_acquisition_mask,
                overwrite=overwrite,
            )
            shapes, datasets = self._build_pyramid(
                group,
                chunks,
                max_layer,
                storage_options,
                overwrite=overwrite,
            )
            self._write_metadata(
                group, max_layer, shapes, datasets, plate_acquisition, well_acquisition
            )

        return plate

    def _write_metadata(
        self, group, max_layer, shapes, datasets, plate_acquisition, well_acquisition
    ):
        coordinate_transformations = well_acquisition.get_coordinate_transformations(
            max_layer=max_layer,
            yx_binning=self._yx_binning,
            ndim=len(shapes[0]),
        )
        fmt = CurrentFormat()
        dims = len(shapes[0])
        fmt.validate_coordinate_transformations(
            dims, len(datasets), coordinate_transformations
        )
        for dataset, transform in zip(
            datasets, coordinate_transformations, strict=True
        ):
            dataset["coordinateTransformations"] = transform
        axes = Axes(well_acquisition.get_axes(), fmt).to_list()
        spatial_calibration_unit = Unit(
            plate_acquisition.get_channel_metadata()[0].spatial_calibration_units
        )
        for axis in axes:
            if axis["name"] in ["z", "y", "x"]:
                axis["unit"] = str(spatial_calibration_unit)
        write_multiscales_metadata(
            group,
            datasets,
            fmt,
            axes,
        )
        group.attrs["omero"] = {
            "channels": plate_acquisition.get_omero_channel_metadata()
        }
        group.attrs["acquisition_metadata"] = {
            "channels": [
                ch_metadata.model_dump()
                for ch_metadata in plate_acquisition.get_channel_metadata().values()
            ]
        }

    def _write_stitched_image(
        self,
        group,
        chunks,
        plate_acquisition,
        storage_options,
        well_acquisition,
        build_acquisition_mask,
        overwrite,
    ):
        stitched_well_da = self._stitch_well_image(
            well_acquisition,
            output_shape=plate_acquisition.get_common_well_shape(),
            build_acquisition_mask=build_acquisition_mask,
        )
        binned_da = self._drop_missing_axes(stitched_well_da, well_acquisition)
        rechunked_da = binned_da.rechunk(self._out_chunks(binned_da.shape, chunks))
        options = self._get_storage_options(storage_options, rechunked_da.shape, chunks)
        wait(
            self._client.persist(
                da.to_zarr(
                    arr=rechunked_da,
                    url=group.store,
                    compute=False,
                    component=str(Path(group.path, "0")),
                    storage_options=options,
                    compressor=options.get(
                        "compressor", zarr.storage.default_compressor
                    ),
                    dimension_separator=group._store._dimension_separator,
                    overwrite=overwrite,
                ),
            )
        )

    def _drop_missing_axes(self, stitched_well_da, well_acquisition):
        drop_axes = tuple(
            i
            for i, s in enumerate(["t", "c", "z", "y", "x"])
            if s not in well_acquisition.get_axes()
        )
        binned_da = self._bin_yx(stitched_well_da)
        if len(drop_axes) > 0:
            binned_da = binned_da.squeeze(drop_axes)
        return binned_da

    def _build_pyramid(
        self,
        group,
        chunks,
        max_layer,
        storage_options,
        overwrite,
    ):
        image = da.from_zarr(url=group.store, component=str(Path(group.path, "0")))
        datasets = [{"path": "0"}]
        shapes = [image.shape]
        for path in range(1, max_layer + 1):
            image = da.coarsen(
                reduction=dask_utils.mean_cast_to(image.dtype),
                x=image,
                axes={
                    image.ndim - 2: 2,
                    image.ndim - 1: 2,
                },
                trim_excess=True,
            )
            options = self._get_storage_options(storage_options, image.shape, chunks)
            image = image.rechunk(options["chunks"])
            wait(
                self._client.persist(
                    da.to_zarr(
                        arr=image,
                        url=group.store,
                        compute=False,
                        component=str(Path(group.path, str(path))),
                        storage_options=options,
                        compressor=options.get(
                            "compressor", zarr.storage.default_compressor
                        ),
                        dimension_separator=group._store._dimension_separator,
                        overwrite=overwrite,
                    )
                )
            )
            datasets.append({"path": str(path)})
            shapes.append(image.shape)
            image = da.from_zarr(
                url=group.store, component=str(Path(group.path, str(path)))
            )

        return shapes, datasets

    def _bin_yx(self, image_da):
        if self._yx_binning > 1:
            return da.coarsen(
                reduction=dask_utils.mean_cast_to(image_da.dtype),
                x=image_da,
                axes={
                    0: 1,
                    1: 1,
                    2: 1,
                    3: self._yx_binning,
                    4: self._yx_binning,
                },
                trim_excess=True,
            )
        return image_da

    def _stitch_well_image(
        self,
        well_acquisition,
        output_shape: tuple[int, int, int, int, int],
        *,
        build_acquisition_mask: bool,
    ):
        from faim_ipa.stitching import DaskTileStitcher

        chunk_shape = well_acquisition.get_tiles()[0].shape

        stitcher = DaskTileStitcher(
            tiles=well_acquisition.get_tiles(),
            chunk_shape=chunk_shape,
            output_shape=output_shape,
            dtype=bool if build_acquisition_mask else well_acquisition.get_dtype(),
        )
        return stitcher.get_stitched_dask_array(
            warp_func=self._warp_func,
            fuse_func=(
                stitching_utils.fuse_sum if build_acquisition_mask else self._fuse_func
            ),
            build_acquisition_mask=build_acquisition_mask,
        )

    def _create_well_group(
        self, plate, well_acquisition, well_sub_group, *, add_to_well_images=True
    ):
        row, col = well_acquisition.get_row_col()
        well_group = plate.require_group(row).require_group(col)
        well_group.require_group(well_sub_group)
        if add_to_well_images:
            zattrs = well_group.attrs.asdict()
            if "well" in zattrs and "images" in zattrs["well"]:
                existing_images = well_group.attrs.asdict()["well"]["images"]
            else:
                existing_images = []
            write_well_metadata(
                well_group, [*existing_images, {"path": well_sub_group}]
            )
        return well_group

    @staticmethod
    def _get_storage_options(
        storage_options: dict,
        output_shape: tuple[int, ...],
        chunks: tuple[int, ...],
    ):
        if storage_options is None:
            return {
                "dimension_separator": "/",
                "compressor": Blosc(cname="zstd", clevel=3, shuffle=Blosc.SHUFFLE),
                "chunks": ConvertToNGFFPlate._out_chunks(output_shape, chunks),
                "write_empty_chunks": False,
            }
        return storage_options

    @staticmethod
    def _out_chunks(shape, chunks):
        if len(shape) == len(chunks):
            return chunks
        return (1,) * (len(shape) - len(chunks)) + chunks



import logging
from pathlib import Path
from typing import List

import numpy as np
from tifffile import imread
from tqdm import tqdm


def average_of_mips(
    image_files: List[Path],
    logger=logging,
):
    """
    Computes a maximum projection of each input image file (assumed to contain a 3D stack), and returns an average of
    all projections.

    Args:
        image_files: list of input image files, assumed to contain a 3D stack each
        logger:

    Returns: an average of maximum projections of all input files (stacks).

    """
    return average_of_thresholded_mips(
        image_files=image_files,
        logger=logger,
    )


def average_of_thresholded_mips(
    image_files: List[Path],
    lower_threshold: int = 0,
    upper_threshold: int = 65535,
    logger=logging,
):
    """
    Computes a maximum projection of each input image file (assumed to contain a 3D stack),
    applies a provided threshold, and returns an average of the thresholded signal of all projections.

    Args:
        image_files: list of input image files, assumed to contain a 3D stack each
        lower_threshold:
        upper_threshold:
        logger:

    Returns: an average of maximum projections of all input files (stacks).

    """

    def image_generator():
        for path in tqdm(sorted(image_files)):
            logger.info(f"Processing {path}.")
            yield imread(path)

    def apply_mask(img, mask):
        img[~mask] = 0
        return img

    images = image_generator()
    image = next(images)
    mask_image = (lower_threshold <= image) & (image <= upper_threshold)

    count_image = np.zeros(mask_image.shape[1:], dtype=np.uint16)
    sum_image = np.max(apply_mask(image, mask_image), axis=0).astype(np.float64)
    count_image += np.max(mask_image, axis=0)

    for image in images:
        mask_image = (lower_threshold <= image) & (image <= upper_threshold)
        sum_image += np.max(apply_mask(image, mask_image), axis=0)
        count_image += np.max(mask_image, axis=0)

    with np.errstate(invalid="ignore"):
        return sum_image / count_image



from faim_ipa.visiview.acquisition import (
    RegionAcquisitionOME,
    RegionAcquisitionSTK,
    StackedTile,
)

__all__ = [
    "RegionAcquisitionOME",
    "RegionAcquisitionSTK",
    "StackedTile",
]



import os
import subprocess
import sys

from faim_ipa.utils import create_logger


def log_commit(task: str = ""):
    command = ["git", "rev-parse", "--verify", "HEAD"]
    info = subprocess.run(command, capture_output=True, text=True, check=False)
    hash_string = info.stdout.strip()

    os.chdir(path=os.getenv("WD", default="."))
    logger = create_logger(name="githash", include_timestamp=False)
    logger.info("[Executing] %s [git-commit] %s", task, hash_string)


if __name__ == "__main__":
    if len(sys.argv) > 1:
        log_commit(task=str(sys.argv[1]))
    else:
        log_commit()



from pathlib import Path

import numpy as np
from numpy._typing import NDArray
from pydantic import BaseModel, NonNegativeInt
from tifffile import imread


class TilePosition(BaseModel):
    time: NonNegativeInt | None
    channel: NonNegativeInt | None
    z: int
    y: int
    x: int

    def __repr__(self):
        return f"TilePosition(time={self.time}, channel={self.channel}, z={self.z}, y={self.y}, x={self.x})"

    def __str__(self):
        return self.__repr__()


class Tile:
    """
    A tile with a path to the image data, shape and position.
    """

    path: str
    shape: tuple[int, ...]
    position: TilePosition
    background_correction_matrix_path: Path | str | None = None
    illumination_correction_matrix_path: Path | str | None = None

    def __init__(
        self,
        path: Path | str,
        shape: tuple[int, int],
        position: TilePosition,
        background_correction_matrix_path: Path | str | None = None,
        illumination_correction_matrix_path: Path | str | None = None,
    ):
        super().__init__()
        self.path = path
        self.shape = shape
        self.position = position
        self.background_correction_matrix_path = background_correction_matrix_path
        self.illumination_correction_matrix_path = illumination_correction_matrix_path

    def __repr__(self):
        return (
            f"Tile(path='{self.path}', shape={self.shape}, "
            f"position={self.position})"
        )

    def __str__(self):
        return self.__repr__()

    def get_yx_position(self) -> tuple[int, int]:
        return self.position.y, self.position.x

    def get_zyx_position(self) -> tuple[int, int, int]:
        return self.position.z, self.position.y, self.position.x

    def get_position(self) -> tuple[int, int, int, int, int]:
        return (
            self.position.time,
            self.position.channel,
            self.position.z,
            self.position.y,
            self.position.x,
        )

    def load_data(self) -> NDArray:
        """
        Load the image data from the path.

        Returns
        -------
        Image data
        """
        data = imread(self.path)
        data = self._apply_background_correction(data)
        return self._apply_illumination_correction(data)

    def load_data_mask(self) -> NDArray:
        """
        Create a binary mask indicating the presence of data.

        Returns
        -------
        Binary mask
        """
        return np.ones(self.shape, dtype=bool)

    def _apply_illumination_correction(self, data):
        dtype = data.dtype
        if self.illumination_correction_matrix_path is not None:
            icm = imread(self.illumination_correction_matrix_path)
            assert icm.shape == data.shape, (
                f"Illumination correction matrix shape {icm.shape} "
                f"does not match image shape {data.shape}."
            )
            mi, ma = np.iinfo(dtype).min, np.iinfo(dtype).max
            data = np.clip(data / icm, a_min=mi, a_max=ma).astype(dtype)
        return data

    def _apply_background_correction(self, data):
        dtype = data.dtype
        if self.background_correction_matrix_path is not None:
            bgcm = imread(self.background_correction_matrix_path)
            assert bgcm.shape == data.shape, (
                f"Background correction matrix shape {bgcm.shape} "
                f"does not match image shape {data.shape}."
            )
            mi, ma = np.iinfo(dtype).min, np.iinfo(dtype).max
            data = np.clip(data - bgcm, a_min=mi, a_max=ma).astype(dtype)
        return data



import numpy as np
from numpy.typing import ArrayLike
from scipy.ndimage import distance_transform_edt


def fuse_random_gradient(
    tiles: ArrayLike, positions: ArrayLike, random_seed=0
) -> ArrayLike:
    """
    Fuses tiles according to positions.
    Where tiles overlap, it chooses a random pixel, weighted according to  a
    linear gradient.
    tiles: ArrayLike, should have shape (tiles, ny, nx)
    positions: ArrayLike, should have shape (tiles, 2)
    """
    np.random.seed(random_seed)
    ny_tot, nx_tot = positions.max(axis=0) + tiles.shape[-2:]
    im_fused = np.zeros((len(tiles), ny_tot, nx_tot), dtype=tiles.dtype)
    im_weight = np.zeros_like(im_fused, dtype="uint16")
    ny_tile, nx_tile = tiles.shape[-2:]

    # distance map to border of image
    mask = np.ones((ny_tile, nx_tile))
    mask[[0, -1], :] = 0
    mask[:, [0, -1]] = 0
    dist_map = distance_transform_edt(mask).astype("uint32") + 1

    for i in range(len(tiles)):
        tile = tiles[i]
        pos = positions[i]
        im_fused[i, pos[0] : pos[0] + ny_tile, pos[1] : pos[1] + nx_tile] += tile
        im_weight[i, pos[0] : pos[0] + ny_tile, pos[1] : pos[1] + nx_tile] += dist_map
    im_weight[:, im_weight.max(axis=0) == 0] = 1
    im_weight = im_weight / im_weight.sum(axis=0)[np.newaxis, :, :]
    im_weight = np.cumsum(im_weight, axis=0)
    im_weight = np.insert(im_weight, 0, np.zeros_like(im_weight[0]), axis=0)
    im_rand = np.random.rand(*im_fused[0].shape)
    for i in range(len(im_fused)):
        im_fused[i, (im_rand < im_weight[i]) | (im_weight[i + 1] < im_rand)] = 0
    return im_fused.sum(axis=0)


def fuse_fw(tiles: ArrayLike, positions: ArrayLike) -> ArrayLike:
    """
    Fuses tiles according to positions.
    Where tiles overlap, it overwrites with the tile that comes later in the
    sequence.
    tiles: ArrayLike, should have shape (tiles, ny, nx)
    positions: ArrayLike, should have shape (tiles, 2)
    """
    ny_tot, nx_tot = positions.max(axis=0) + tiles.shape[-2:]
    im_fused = np.zeros((ny_tot, nx_tot), dtype=tiles.dtype)
    ny_tile, nx_tile = tiles.shape[-2:]

    for tile, pos in zip(tiles, positions, strict=True):
        im_fused[pos[0] : pos[0] + ny_tile, pos[1] : pos[1] + nx_tile] = tile

    return im_fused


def fuse_rev(tiles: ArrayLike, positions: ArrayLike) -> ArrayLike:
    """
    Fuses tiles according to positions.
    Where tiles overlap, it overwrites with the tile that comes earlier in the
    sequence.
    tiles: ArrayLike, should have shape (tiles, ny, nx)
    positions: ArrayLike, should have shape (tiles, 2)
    """
    ny_tot, nx_tot = positions.max(axis=0) + tiles.shape[-2:]
    im_fused = np.zeros((ny_tot, nx_tot), dtype=tiles.dtype)
    ny_tile, nx_tile = tiles.shape[-2:]

    for tile, pos in reversed(list(zip(tiles, positions, strict=True))):
        im_fused[pos[0] : pos[0] + ny_tile, pos[1] : pos[1] + nx_tile] = tile

    return im_fused
